# -*- coding: utf-8 -*-
"""
Created on Thu Nov 12 16:19:28 2015

@author: wangronin
"""

import pdb
import numpy as np
from numpy import pi, log

## SMSE measurement
# test_y is the target, pred_y the predicted target, both 1D arrays of same length
def SMSE(test_y,pred_y):
	se = []
	target_variance = np.var(test_y)
	for i in range(len(test_y)):
		temp = (pred_y[i] - test_y[i])**2
		se.append(temp)
	mse = np.mean(se)
	smse = mse / target_variance
	return smse

## MSLL = mean standardized log loss
## logprob = 0.5*log(2*pi.*varsigmaVec) + sserror - 0.5*log(2*pi*varyTrain)...
##           - ((yTestVec - meanyTrain).^2)./(2*varyTrain);
def MSLL(train_y,test_y,pred_y,variances):
	sll = []
	mean_y = np.mean(train_y)
	var_y = np.var(train_y)
	for i in range(len(variances)):
		if variances[i] == 0:
			variances[i] += 0.0000001 #hack
		sll_trivial = 0.5*log(2 * pi * var_y) + ((test_y[i] - mean_y)**2 / (2* var_y))
		sllv = ( 0.5*log(2 * pi * variances[i]) + \
      ((test_y[i] - pred_y[i])**2 / (2* variances[i])) ) - sll_trivial
		sll.append(sllv)
	sll = np.array(sll)
	msll = np.mean(sll)
	return msll

# # Obtain the initial design locations
# def get_design_sites(dim, n_sample, x_lb, x_ub, sampling_method='lhs'):

#     x_lb = atleast_2d(x_lb)
#     x_ub = atleast_2d(x_ub)

#     x_lb = x_lb.T if size(x_lb, 0) != 1 else x_lb
#     x_ub = x_ub.T if size(x_ub, 0) != 1 else x_ub

#     if sampling_method == 'lhs':
#         # Latin Hyper Cube Sampling: Get evenly distributed sampling in R^dim
#         samples = lhs(dim, samples=n_sample) * (x_ub - x_lb) + x_lb

#     elif sampling_method == 'uniform':
#         samples = np.random.rand(n_sample, dim) * (x_ub - x_lb) + x_lb

#     elif sampling_method == 'sobol':
#         seed = mod(int(time.time()) + os.getpid(), int(1e6))
#         samples = np.zeros((n_sample, dim))
#         for i in range(n_sample):
#             samples[i, :], seed = i4_sobol(dim, seed)
#         samples = samples * (x_ub - x_lb) + x_lb

#     elif sampling_method == 'halton':
#         sequencer = Halton(dim)
#         samples = sequencer.get(n_sample) * (x_ub - x_lb) + x_lb

#     return samples

# diagostic plots of gradient field
# def plot_contour_gradient(ax, f, grad, x_lb, x_ub, title='f', is_log=False, n_level=30, foo=0,
#                          f_data=None, grad_data=None, n_per_axis=200):
#    fig = ax.figure

#    x = np.linspace(x_lb[0], x_ub[0], n_per_axis)
#    y = np.linspace(x_lb[1], x_ub[1], n_per_axis)
#    X, Y = np.meshgrid(x, y)

# #    divider = make_axes_locatable(ax)
# #    cax = divider.append_axes("right", size="5%", pad=0.05)

#    if f_data is None:
#        fitness = np.array([f(p.reshape(1, -1)) for p in np.c_[X.flatten(), Y.flatten()]]).reshape(-1, len(x))
#    else:
#        fitness = f_data
# #       fitness = (fitness - np.min(fitness)) / (np.max(fitness) - np.min(fitness)) + foo
#    if is_log:
#        fitness = np.log(fitness)
       
#    CS = ax.contour(X, Y, fitness, n_level, cmap=plt.cm.Spectral, linewidths=1)
#    plt.clabel(CS, inline=1, fontsize=10)
       
#    if grad is not None:
#        # calculate function gradients
#        x1 = np.linspace(x_lb[0], x_ub[0], np.floor(n_per_axis / 10))
#        x2 = np.linspace(x_lb[1], x_ub[1], np.floor(n_per_axis / 10))
#        X1, X2 = np.meshgrid(x1, x2)
#        if grad_data is None:
#            dx = np.array([grad(p.reshape(1, -1)).flatten() for p in np.c_[X1.flatten(), X2.flatten()]])
#            np.save('grad.npy', dx)
#        else:
#            dx = grad_data
           
#        dx_norm = np.sqrt(np.sum(dx ** 2.0, axis=1)) # in case of zero gradients
#        dx /= dx_norm.reshape(-1, 1)
#        dx1 = dx[:, 0].reshape(-1, len(x1))
#        dx2 = dx[:, 1].reshape(-1, len(x1))
       
#        CS = ax.quiver(X1, X2, dx1, dx2, dx_norm, cmap=plt.cm.jet,
#                       #norm=colors.LogNorm(vmin=1e-100, vmax=dx_norm.max()),
#                       headlength=5)

# #        fig.colorbar(CS, ax=ax)
    
#    ax.set_xlabel('$x_1$')
#    ax.set_ylabel('$x_2$')
#    ax.grid(True)
#    ax.set_title(title)
#    ax.set_xlim(x_lb[0], x_ub[0])
#    ax.set_ylim(x_lb[1], x_ub[1])
   
   
# def plot_surface_contour(ax, f, grad, x_lb, x_ub, title='f', 
#                          log_transform=False, n_level=30, foo=0,
#                          f_data=None, grad_data=None,n_per_axis=200):
    
#     fig = ax.figure
#     x = np.linspace(x_lb[0], x_ub[0], n_per_axis)
#     y = np.linspace(x_lb[1], x_ub[1], n_per_axis)
#     X, Y = np.meshgrid(x, y)

# #    divider = make_axes_locatable(ax)
# #    cax = divider.append_axes("right", size="5%", pad=0.05)

#     if f_data is None:
#         fitness = np.array([f(p.reshape(1, -1)) for p in np.c_[X.flatten(), Y.flatten()]]).reshape(-1, len(x))
#     else:
#         fitness = f_data
#     try:
#         fitness = (fitness - np.min(fitness)) / (np.max(fitness) - np.min(fitness)) + foo
#         if log_transform:
#             fitness = np.log(fitness)
#         CS = ax.contour(X, Y, fitness, n_level, cmap=plt.cm.winter, linewidths=1, offset=0)
#         plt.clabel(CS, inline=1, fontsize=15)
        
#         tri = mtri.Triangulation(X, Y)
#         ax.plot_trisurf(X, Y, fitness, rstride=1, cstride=1, cmap=plt.cm.Spectral, linewidth=0, alpha=0.3)
# #        fig.colorbar(CS, ax=ax, fraction=0.046, pad=0.04)
#     except:
#         pdb.set_trace()

#    # calculate function gradients
#     x1 = np.linspace(x_lb[0], x_ub[0], np.floor(n_per_axis / 10))
#     x2 = np.linspace(x_lb[1], x_ub[1], np.floor(n_per_axis / 10))
#     X1, X2 = np.meshgrid(x1, x2)

#     if grad is not None:
#         if grad_data is None:
#             dx = np.array([grad(p).flatten() for p in np.c_[X1.flatten(), X2.flatten()]])
#             np.save('grad.npy', dx)
#         else:
#             dx = grad_data

#         dx_norm = np.sqrt(np.sum(dx ** 2.0, axis=1)) # in case of zero gradients
#         dx /= dx_norm.reshape(-1, 1)
#         dx1 = dx[:, 0].reshape(-1, len(x1))
#         dx2 = dx[:, 1].reshape(-1, len(x1))

#         CS = ax.quiver(X1, X2, dx1, dx2, dx_norm, cmap=plt.cm.jet,
#                       #norm=colors.LogNorm(vmin=1e-100, vmax=dx_norm.max()),
#                       headlength=5)

# #        fig.colorbar(CS, ax=ax)
    
#     ax.set_xlabel('$x_1$')
#     ax.set_ylabel('$x_2$')
#     ax.grid(True)
#     ax.set_title(title)
#     ax.set_xlim(x_lb[0], x_ub[0])
#     ax.set_ylim(x_lb[1], x_ub[1])


import os, logging
import numpy as np
from typing import Callable, Any, Tuple, List, Union, Optional

from . import AcquisitionFunction, Surrogate
from .BayesOpt import BO, ParallelBO, NoisyBO, AnnealingBO
from .Solution import Solution
from .Surrogate import RandomForest, GaussianProcess, trend
from .SearchSpace import SearchSpace, OrdinalSpace, ContinuousSpace, NominalSpace
from .Extension import OptimizerPipeline

__all__ = [
    'BO', 'ParallelBO', 'NoisyBO', 'AnnealingBO', 'Solution', 'AcquisitionFunction', 
    'Surrogate', 'RandomForest', 'GaussianProcess', 'trend',
    'SearchSpace', 'OrdinalSpace', 'ContinuousSpace', 
    'NominalSpace', 'RandomForest', 'OptimizerPipeline', 'fmin'
]

# To use `dill` for the pickling, which works for
# much more python objects
os.environ['LOKY_PICKLER'] = 'dill' 

verbose = {
    False : logging.NOTSET,
    'DEBUG' : logging.DEBUG,
    'INFO' : logging.INFO
}

Vector = List[float]
Matrix = List[Vector]

# TODO: implement the code for `callback`, `xtol`, `ftol`, and `kwargs`
def fmin(
    func: Callable, 
    lower: Union[float, Vector],
    upper: Union[float, Vector],
    x0: Union[int, Matrix, np.ndarray, None] = None, 
    y0: Union[Vector, None] = None, 
    n_point: int = 1,
    args: Tuple = (), 
    xtol: float = 1e-4, 
    ftol: float = 1e-4,
    max_FEs: Optional[int] = None,
    verbose: Optional[bool] = False,
    callback: Optional[Callable] = None,
    seed: Optional[int] = None,
    **kwargs
    ) -> Tuple[Vector, float, int, int, List[np.ndarray]]:
    """ Minimize a function using the Bayesian Optimization algorithm, which only uses 
    function values, not derivatives or second derivatives. This function maintains an 
    interface similar to `scipy.optimize.fmin`. Hereafter, we use the following customized 
    types to describe the usage:

    - Vector = List[float]
    - Matrix = List[Vector]

    Parameters
    ----------
    func : Callable 
        The objective function to be minimized.
    lower : Union[float, Vector]
        The lower bound of search variables, from which the search dimension is inferred. 
        When it is not a `float`, it must have the same length as `upper`.
    upper : Union[float, Vector]
        The upper bound of search variables, from which the search dimension is inferred. 
        When it is not a `float`, it must have the same length as `lower`.
    x0 : Union[int, Matrix, None], optional
        When it takes integer values, it specifies the number of initial sample points to 
        take; When it is a 2d numpy array (or a nested list that contains the same data), 
        it provides the initial sample points.
    y0 : Union[Vector, None], optional
        When `x0` is a 2d numpy array, it contains the function values pertaining to `x0`;
        Otherwise, it is ignored. It is by default None. 
    n_point : int, optional
        The number of trial points generated in each iteration, by default 1
    args : Tuple, optional
        Extra arguments passed to `func`, i.e., ``func(x, *args)``.
    xtol : float, optional
        Absolute error in xopt between iterations that is acceptable for convergence, 
        by default 1e-4.
    ftol : float, optional
        Absolute error in func(xopt) between iterations that is acceptable for convergence, 
        by default 1e-4.
    max_FEs : Optional[int], optional
        Maximal number of function evaluations to make, by default None.
    verbose : Optional[bool], optional
        Verbosity of the output, by default False.
    callback : Optional[Callable], optional
        Called after each iteration, as `callback(X)`, where `X` is the current parameter 
        vectors, by default None.
    seed : Optional[int], optional
        Seeding the random number generator in `numpy`, by default None.

    Returns
    -------
    Tuple[Vector, float, int, int, List[np.ndarray]]
        The return value contains:
        - The best-so-far parameter vector
        - The best-so-far function value
        - The number of iterations
        - The number of function evaluations
        - A list of trial points/vectors generated in every iteration

    Examples
    --------
    >>> def f(x):
    ...     return sum(x ** 2)
    >>> from bayes_optim import fmin
    >>> 
    >>> minimum = fmin(f, [-5] * 2, [5] * 2, max_FEs=30, seed=42)
    Optimization terminated successfully.
            Current function value: 0.007165794451494286
            Iterations: 21
            Function evaluations: 30
    >>> minimum[0]
    0.007165794451494286
    """
    if seed is not None:
        _state = np.random.get_state()
        np.random.seed(seed)

    obj_func = lambda x: func(x, *args)
    
    if isinstance(lower, float) and isinstance(upper, float):
        dim = 1
        space = ContinuousSpace([lower, upper])
    elif isinstance(lower, list) and isinstance(upper, list):
        assert len(lower) == len(upper)
        dim = len(lower)
        space = ContinuousSpace(list(zip(lower, upper))) 
        lower, upper = np.array(lower), np.array(upper)

    mean = trend.constant_trend(dim, beta=0)
    thetaL = 1e-10 * (upper - lower) * np.ones(dim)
    thetaU = 10 * (upper - lower) * np.ones(dim)
    model = GaussianProcess(
        mean=mean, corr='squared_exponential',
        thetaL=thetaL, thetaU=thetaU,
        nugget=0, noise_estim=False,
        optimizer='BFGS', wait_iter=3, random_start=dim,
        likelihood='concentrated', eval_budget=100 * dim
    )

    # set up the warm-starting and DoE size
    if isinstance(x0, int):
        DoE_size = x0
        warm_data = ()
    elif hasattr(x0, '__iter__'): 
        DoE_size = None 
        if y0 is None:
            y0 = [obj_func(_) for _ in x0]
        warm_data = (x0, y0)
    else:
        DoE_size = None 
        warm_data = ()
    
    _BO = BO if n_point == 1 else ParallelBO
    opt = _BO(
        search_space=space, 
        obj_fun=obj_func, 
        model=model, 
        DoE_size=DoE_size,
        warm_data=warm_data,
        eval_type='list',
        max_FEs=max_FEs, 
        verbose=verbose, 
        n_point=n_point
    )
    opt.run()

    if seed is not None:
        np.random.set_state(_state)

    N, n = opt.DoE_size, opt.n_point
    _data, data = opt.data[:N, :], opt.data[N:, :]

    data_per_iteration = [np.asarray(data[:N, :])]
    data_per_iteration += [
        np.asarray(data[(i * n):((i + 1) * n), :]) for i in range(opt.iter_count - 1)
    ]

    print(
        "Optimization terminated successfully.\n"\
        "        Current function value: {}\n"\
        "        Iterations: {}\n"\
        "        Function evaluations: {}\n".format(
            opt.fopt, opt.iter_count, opt.eval_count
        )
    )
    return opt.xopt, opt.fopt, opt.iter_count, opt.eval_count, data_per_iteration



from pdb import set_trace
import logging, sys
import numpy as np

from scipy.linalg import solve_triangular
from typing import Callable, Any, Tuple, List, Union
from ..utils import dynamic_penalty
from ..misc import handle_box_constraint, LoggerFormatter

Vector = List[float]
Matrix = List[Vector]

def set_bounds(bound, dim):
    if isinstance(bound, str):
        bound = eval(bound)
    elif isinstance(bound, (float, int)):
        bound = [bound] * dim
    elif hasattr(bound, '__iter__'):
        bound = list(bound)
        if len(bound) == 1:
            bound *= dim
    assert len(bound) == dim
    return np.asarray(bound)

class OnePlusOne_CMA(object):
    def __init__(
        self, 
        dim: int, 
        obj_fun: Callable,
        h: Callable = None,
        g: Callable = None,
        x0: Union[str, Vector, np.ndarray] = None,
        sigma0: Union[float] = None, 
        C0: Union[Matrix, np.ndarray] = None, 
        lb: Union[float, str, Vector, np.ndarray] = -np.inf,
        ub: Union[float, str, Vector, np.ndarray] = np.inf,
        ftarget: Union[int, float] = None,
        max_FEs: Union[int, str] = np.inf, 
        minimize: bool = True,
        xtol: float = 1e-4, 
        ftol: float = 1e-4,
        verbose: bool = False,
        logger: str = None,
        **kwargs
        ):
        """ (1+1)-CMA-ES and (1+1)-Cholesky-CMA-ES 
        Hereafter, we use the following customized 
        types to describe the usage:

        - Vector = List[float]
        - Matrix = List[Vector]

        Parameters
        ----------
        dim : int
            Dimensionality of the search space.
        obj_fun : Callable
            The objective function to be minimized.
        h : Callable, optional
            The equality constraint function, by default None.
        g : Callable, optional
            The inequality constraint function, by default None.
        x0 : Union[str, Vector, np.ndarray], optional
            The initial guess (by default None) which must fall between lower and 
            upper bounds, if non-infinite values are provided for `lb` and `ub`. 
            Note that, `x0` must be provided when `lb` and `ub` both take infinite values.
        sigma0 : Union[float], optional
            The initial step size, by default None
        C0 : Union[Matrix, np.ndarray], optional
            The initial covariance matrix which must be positive definite, by default None.
            Any non-positive definite input will be ignored.
        lb : Union[float, str, Vector, np.ndarray], optional
            The lower bound of search variables. When it is not a `float`, it must have 
            the same length as `upper`, by default `-np.inf`.
        ub : Union[float, str, Vector, np.ndarray], optional
            The upper bound of search variables. When it is not a `float`, it must have 
            the same length as `lower`, by default `np.inf`.
        ftarget : Union[int, float], optional
            The target value to hit, by default None.
        max_FEs : Union[int, str], optional
            Maximal number of function evaluations to make, by default `np.inf`.
        minimize : bool, optional
            To minimize or maximize, by default True.
        xtol : float, optional
            Absolute error in xopt between iterations that is acceptable for convergence, 
            by default 1e-4.
        ftol : float, optional
            Absolute error in func(xopt) between iterations that is acceptable for convergence, 
            by default 1e-4.
        verbose : bool, optional
            Verbosity of the output, by default False.
        logger : str, optional
            Name of the logger file, by default None, which turns off the logging behaviour.
        """
        self.dim = dim
        self.obj_fun = obj_fun
        self.h = h
        self.g = g
        self.minimize = minimize
        self.ftarget = ftarget
        self.lb = set_bounds(lb, self.dim)
        self.ub = set_bounds(ub, self.dim)
        self.sigma0 = self.sigma = sigma0

        self.x = x0
        self.max_FEs = int(eval(max_FEs)) if isinstance(max_FEs, str) else max_FEs
        self._better = lambda a, b: a <= b if self.minimize else lambda a, b: a >= b
        self._init_aux_var(kwargs)
        self._init_covariance(C0)

        self.eval_count = 0
        self.iter_count = 0
        self.xopt = self._x
        self.fopt = np.inf
        self.__fopt = np.inf   # penalized `fopt`
        self.stop_dict = {}
        self._exception = False
        self.verbose = verbose
        self.logger = logger
        
        # parameters for stopping criteria
        self.xtol = xtol
        self.ftol = ftol
        self._delta_x = None 
        self._delta_f = None
        
    def _init_aux_var(self, opts):
        self.prob_target = opts['p_succ_target'] if 'p_succ_target' in opts else 2 / 11
        self.threshold = opts['p_threshold'] if 'p_threshold' in opts else 0.44
        self.d = opts['d'] if 'd' in opts else 1 + self.dim / 2
        self.ccov = opts['ccov'] if 'ccov' in opts else 2 / (self.dim ** 2 + 6)
        self.cp = opts['cp'] if 'cp' in opts else 1 / 12
        self.cc = opts['cc'] if 'cc' in opts else 2 / (self.dim + 2)

        self.success_rate = self.prob_target
        self.pc = np.zeros(self.dim)
        self._coeff = self.cc * (2 - self.cc)

    def _init_covariance(self, C):
        if C is None:
            self._C = np.eye(self.dim)
            self._A = np.eye(self.dim)
        else:
            self.C = C

    @property
    def logger(self):
        return self._logger

    @logger.setter
    def logger(self, logger):
        if isinstance(logger, logging.Logger):
            self._logger = logger
            self._logger.propagate = False
            return

        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.setLevel(logging.DEBUG)
        fmt = LoggerFormatter()

        if self.verbose:
            # create console handler and set level to warning
            ch = logging.StreamHandler(sys.stdout)
            ch.setLevel(logging.INFO)
            ch.setFormatter(fmt)
            self._logger.addHandler(ch)

        # create file handler and set level to debug
        if logger is not None:
            fh = logging.FileHandler(logger)
            fh.setLevel(logging.DEBUG)
            fh.setFormatter(fmt)
            self._logger.addHandler(fh)

        if hasattr(self, 'logger'):
            self._logger.propagate = False

    @property
    def C(self):
        return self._C
    
    @C.setter
    def C(self, C):
        if C is not None:
            try:
                A = np.linalg.cholesky(C)
                if np.all(np.isreal(A)):
                    # TODO: `_A` should be a private attribute
                    self._A = A
                    self._C = C
            except np.linalg.LinAlgError:
                pass

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, x):
        if x is not None:
            x = eval(x) if isinstance(x, str) else x
            x = np.asarray(x)
            assert np.all(x - self.lb >= 0)
            assert np.all(x - self.ub <= 0)
        else:
            assert all(~np.isinf(self.lb)) & all(~np.isinf(self.ub))
            x = (self.ub - self.lb) * np.random.rand(self.dim) + self.lb
        self._x = x
        
    @property
    def sigma(self):
        return self._sigma
    
    @sigma.setter
    def sigma(self, sigma):
        if sigma is None:
            assert all(~np.isinf(self.lb)) & all(~np.isinf(self.ub))
            sigma = np.max(self.ub - self.lb) / 5
        assert sigma > 0
        self._sigma = sigma

    def run(self):
        while not self.check_stop():
            self.step()
        return self.xopt, self.fopt, self.stop_dict
            
    def step(self):
        x = self.ask()
        y = self.evaluate(x)
        self.tell(x, y)
    
    def evaluate(self, x):
        return self.obj_fun(x)

    def ask(self, n_point=1) -> np.ndarray:
        """The mutation function

        Parameters
        ----------
        n_point : int, optional
            The number of mutants, which is always 1. This argument is only meant to keep
            the function interface consistant.

        Returns
        -------
        np.ndarray
            The mutation vector
        """
        z = np.random.randn(self.dim)
        x = self._x + self.sigma * z.dot(self._A.T)
        x = handle_box_constraint(x, self.lb, self.ub) 
        return x

    def tell(self, x, y):
        if hasattr(y, '__iter__'):
            y = y[0]

        _y = y + dynamic_penalty(
            x, self.iter_count + 1, 
            self.h, self.g, 
            minimize=self.minimize
        )
            
        success = self._better(_y, self.__fopt)
        z = (x - self._x) / self._sigma
        self._update_step_size(success)

        if success:
            self._delta_f = self.__fopt - _y
            self._delta_x = np.sqrt(np.sum((self.xopt - x) ** 2))
            self.fopt = y
            self.__fopt = _y
            self._x = self.xopt = x
            self._update_covariance(z)

        self._handle_exception()
        self.eval_count += 1
        self.iter_count += 1

        if self.verbose:
            self._logger.info('iteration {},'.format(self.eval_count))
            self._logger.info('fopt: {}'.format(self.fopt)) 
            self._logger.info('sigma: {}'.format(self._sigma)) 
            self._logger.info('xopt: {}\n'.format(self.xopt.tolist()))

    def check_stop(self):
        if self.ftarget is not None:
            if self._better(self.fopt, self.ftarget):
                self.stop_dict['ftarget'] = self.fopt
            
        if self.eval_count >= self.max_FEs:
            self.stop_dict['FEs'] = self.eval_count
        
        # TODO: add this as an option: lower and upper bounds for regular sigmas
        if self.sigma < 1e-8 or self.sigma > 1e8:
            self.stop_dict['sigma'] = self.sigma

        if self._delta_f is not None and self._delta_f < self.ftol:
            self.stop_dict['ftol'] = self._delta_f
        
        if self._delta_x is not None and self._delta_x < self.xtol:
            self.stop_dict['xtol'] = self._delta_x

        return bool(self.stop_dict)

    def _update_covariance(self, z):
        if self.success_rate < self.threshold:
            self.pc = (1 - self.cc) * self.pc + np.sqrt(self._coeff) * z
            self._C = (1 - self.ccov) * self._C + self.ccov * np.outer(self.pc, self.pc)
        else:
            self.pc = (1 - self.cc) * self.pc
            self._C = (1 - self.ccov * (1 - self._coeff)) * self._C + \
                self.ccov * np.outer(self.pc, self.pc)

        self._C = np.triu(self._C) + np.triu(self._C, 1).T 
        self._update_A(self._C)

    def _update_step_size(self, success):
        self.success_rate = (1 - self.cp) * self.success_rate + self.cp * success
        self._sigma *= np.exp(
            (self.success_rate - self.prob_target) / (1 - self.prob_target) / self.d
        )
    
    def _update_A(self, C):
        if np.any(np.isinf(C)):
            self._exception = True
        else:
            try:
                A = np.linalg.cholesky(C)
                if np.any(~np.isreal(A)):
                    self._exception = True
                else:
                    self._A = A
            except np.linalg.LinAlgError:
                self._exception = True

    def _handle_exception(self):
        if self._sigma < 1e-8 or self._sigma > 1e8:
            self._exception = 1
        
        if self._exception:
            self._C = np.eye(self.dim)
            self.pc = np.zeros((self.dim, 1))
            self._A = np.eye(self.dim)
            self._sigma = self.sigma0
            self._exception = False

class OnePlusOne_Cholesky_CMA(OnePlusOne_CMA):
    def _init_covariance(self, C):
        reset = False
        if C is not None:
            try:
                A = np.linalg.cholesky(C)
                if np.any(~np.isreal(A)):
                    reset = True
                else:
                    self.A = A
            except np.linalg.LinAlgError:
                reset = True

        if C is None or reset:
            self.A = np.eye(self.dim)
        
    @property
    def A(self):
        return self._A
    
    @A.setter
    def A(self, A):
        assert np.all(np.triu(A, k=1).ravel() == 0)
        self._A = A
        self._A_inv = solve_triangular(A, np.eye(self.dim), lower=True)

    def _update_covariance(self, z):
        cb = self.ccov
        if self.success_rate < self.threshold:
            self.pc = (1 - self.cc) * self.pc + np.sqrt(self._coeff) * z
            ca = (1 - self.ccov)
        else:
            self.pc = (1 - self.cc) * self.pc
            ca = (1 - self.ccov) + self.ccov * self.cc * (2 - self.cc)

        w = self.pc.dot(self._A_inv.T)
        w_ = w.dot(self._A_inv)
        L = np.sum(w ** 2)
        
        self._A += (np.sqrt(1 + L * cb / ca) - 1) / L * np.outer(self.pc, w)
        self._A *= np.sqrt(ca)

        self._A_inv -= (1 - 1 / np.sqrt(1 + L * cb / ca)) / L * np.outer(w, w_)
        self._A_inv *= 1 / np.sqrt(ca)

    def _handle_exception(self):
        pass


"""
Created on Wed Aug 23 16:48:47 2017

@author: Hao Wang
@email: wangronin@gmail.com
"""

from pdb import set_trace
import numpy as np

from abc import abstractmethod, ABC
from numpy import newaxis, zeros, tile, eye, r_, c_, ones, array, atleast_2d
from sklearn.ensemble import RandomForestRegressor

class Trend(ABC):
    pass

class BasisExpansionTrend(Trend):
    def __init__(self, n_feature, n_dim=None, beta=None):
        """n_dim : the dimension of the function space of the trend function
        """
        self.n_feature = int(n_feature)
        self.n_dim = int(n_dim)
        self.beta = beta

    @property
    def beta(self):
        return self._beta

    @beta.setter
    def beta(self, beta):
        if beta is not None:
            if not hasattr(beta, '__iter__'):
                beta = array([beta] * self.n_dim)
            beta = atleast_2d(beta).reshape(-1, 1)
            if len(beta) != self.n_dim:
                raise Exception('Shapes of beta and F do not match.')
        self._beta = beta

    def __str__(self):
        return self.__class__

    def __call__(self, X):
        if self._beta is None:
            raise Exception('beta is not set!')
        return self.F(X).dot(self._beta)
    
    @abstractmethod
    def F(self, X):
        "Evaluate the function basis as X"
    
    @abstractmethod
    def Jacobian(self, X):
        "Compute the Jacobian matrix of function basis"

    @abstractmethod
    def Hessian(self, X):
        "Compute the Hessian tensor of function basis"

    def check_input(self, X):
        # Check input shapes
        X = np.atleast_2d(X)
        if X.shape[1] != self.n_feature:
            X = X.T
        if X.shape[1] != self.n_feature:
            raise Exception('x does not have the right size!')
        return X

    def __eq__(self, trend_b):
        pass

    def __add__(self, trend_b):
        pass

# TODO: change all Jacobian function to denominator layout
# TODO: also to unify the naming of classes.
class constant_trend(BasisExpansionTrend):
    """Zero order polynomial (constant, p = 1) regression model.

    x --> f(x) = 1

    """
    def __init__(self, n_feature, beta=None):
        super(constant_trend, self).__init__(n_feature, 1, beta)

    def F(self, X):
        X = self.check_input(X)
        n_eval = X.shape[0]
        return ones((n_eval, 1))

    def Jacobian(self, x):
        self.check_input(x)
        return zeros((1, self.n_feature)) # numerator layout

    def Hessian(self, x):
        self.check_input(x)
        # TODO: unified the layout for matrix calculus here!!!
        return zeros((self.n_feature, self.n_feature, self.n_dim)) # denominator layout

class linear_trend(BasisExpansionTrend):
    """First order polynomial (linear, p = n+1) regression model.

    x --> f(x) = [ 1, x_1, ..., x_n ].T
    """
    def __init__(self, n_feature, beta=None):
        super(linear_trend, self).__init__(n_feature, n_feature + 1, beta)
    
    def F(self, X):
        X = self.check_input(X)
        n_eval = X.shape[0]
        return c_[ones(n_eval), X]

    def Jacobian(self, x):
        x = self.check_input(x)
        assert x.shape[0] == 1
        return r_[zeros((1, self.n_feature)), eye(self.n_feature)]
    
    def Hessian(self, x):
        self.check_input(x)
        # TODO: unified the layout for matrix calculus here!!!
        return zeros((self.n_feature, self.n_feature, self.n_dim)) # denominator layout

class quadratic_trend(BasisExpansionTrend):
    """
    Second order polynomial (quadratic, p = n * (n-1) / 2 + n + 1) regression model.

    x --> f(x) = [ 1, { x_i, i = 1,...,n }, { x_i * x_j,  (i,j) = 1,...,n } ].T
                                                          i > j
    """
    def __init__(self, n_feature, beta=None):
        super(quadratic_trend, self).__init__(
            n_feature, (n_feature + 1) * (n_feature + 2) / 2, beta
        )

    def F(self, X):
        X = self.check_input(X)
        n_eval = X.shape[0]
        f = c_[ones(n_eval), X]
        for k in range(self.n_feature):
            f = c_[f, X[:, k, np.newaxis] * X[:, k:]]
        return f

    def Jacobian(self, X):
        raise NotImplementedError

    def Hessian(self, X):
        raise NotImplementedError

class NonparametricTrend(Trend):
    def __init__(self, X, y):
        self.regr = RandomForestRegressor(20)
        self.regr.fit(X, y)

    def __call__(self, X):
        return self.regr.predict(X)

if __name__ == '__main__':
    T = linear_trend(2, beta=(1, 2, 10))

    X = np.random.randn(5, 2)
    print(T(X))
    print(T.Jacobian(X))


# TODO: remove those functions
# legacy functions
def constant(x):

    """
    Parameters
    ----------
    x : array_like
        An array with shape (n_eval, n_features) giving the locations x at
        which the regression model should be evaluated.

    Returns
    -------
    f : array_like
        An array with shape (n_eval, p) with the values of the regression
        model.
    """
    x = np.asarray(x, dtype=np.float64)
    n_eval = x.shape[0]
    f = np.ones([n_eval, 1])
    return f


def linear(x):
    """
    Parameters
    ----------
    x : array_like
        An array with shape (n_eval, n_features) giving the locations x at
        which the regression model should be evaluated.

    Returns
    -------
    f : array_like
        An array with shape (n_eval, p) with the values of the regression
        model.
    """
    x = np.asarray(x, dtype=np.float64)
    n_eval = x.shape[0]
    f = np.hstack([np.ones([n_eval, 1]), x])
    return f


def quadratic(x):
    """
    Second order polynomial (quadratic, p = n*(n-1)/2+n+1) regression model.

    x --> f(x) = [ 1, { x_i, i = 1,...,n }, { x_i * x_j,  (i,j) = 1,...,n } ].T
                                                          i > j

    Parameters
    ----------
    x : array_like
        An array with shape (n_eval, n_features) giving the locations x at
        which the regression model should be evaluated.

    Returns
    -------
    f : array_like
        An array with shape (n_eval, p) with the values of the regression
        model.
    """

    x = np.asarray(x, dtype=np.float64)
    n_eval, n_features = x.shape
    f = np.hstack([np.ones([n_eval, 1]), x])
    for k in range(n_features):
        f = np.hstack([f, x[:, k, np.newaxis] * x[:, k:]])

    return f



import json
from copy import copy, deepcopy

import numpy as np
from numpy.random import randint, rand

from abc import abstractmethod
from pyDOE import lhs
from scipy.special import logit

# TODO: rename `sampling` --> `sample`
# TODO: add conditional parameters

TRANS = {
    'log': np.log,
    'log10': np.log10,
    'logit': logit,
    'bilog': lambda x: np.sign(x) * np.log(1 + np.abs(x))
}

INV_TRANS = {
    'log': np.exp,
    'log10': lambda x: np.power(10, x),
    'logit': lambda x: 1 / (1 + np.exp(-x)),
    'bilog': lambda x: np.sign(x) * (np.exp(np.abs(x)) - 1) 
}

class SearchSpace(object):
    def __init__(self, bounds, var_name, name, random_seed=None):
        """Search Space Base Class

        Parameters
        ----------
        bounds : (list of) list,
            lower and upper bound for continuous/ordinal parameter type
            categorical values for nominal parameter type.
            The dimension of the space is determined by the length of the 
            nested list
        var_name : (list of) str,
            variable name per dimension. If only a string is given for multiple 
            dimensions, variable names are created by appending counting numbers
            to the input string. 
        name : str,
            search space name. It is typically used as the grouping variable
            when converting the Solution object to dictionary, allowing for 
            vector-valued search parameters. See 'to_dict' method below.

        Attributes
        ----------
        dim : int,
            dimensinality of the search space
        bounds : a list of lists,
            each sub-list stores the lower and upper bound for continuous/ordinal variable
            and categorical values for nominal variable
        levels : a list of lists,
            each sub-list stores the categorical levels for every nominal variable. It takes
            `None` value when there is no nomimal variable
        precision : a list of double,
            the numerical precison (granularity) of continuous parameters, which usually  
            very practical in real-world applications
        var_name : a list of str,
            variable names per dimension 
        var_type : a list of str, 
            variable type per dimension, 'C': continuous, 'N': nominal, 'O': ordinal
        C_mask : a bool array,
            the mask array for continuous variables
        O_mask : a bool array,
            the mask array for integer variables
        N_mask : a bool array,
            the mask array for discrete variables
        id_C : an int array,
            the index array for continuous variables
        id_O : an int array,
            the index array for integer variables
        id_N : an int array,
            the index array for discrete variables
        """
        if hasattr(bounds[0], '__iter__') and not isinstance(bounds[0], str):
            self.bounds = [tuple(b) for b in bounds]
        else:
            self.bounds = [tuple(bounds)]
            
        self.dim = len(self.bounds)
        self.name = name
        self.random_seed = random_seed
        self.var_type = None 
        self.levels = None
        self.precision = {}
        self.scale = {}

        if var_name is not None:
            if isinstance(var_name, str):
                if self.dim > 1:
                    var_name = [var_name + '_' + str(_) for _ in range(self.dim)]
                else:
                    var_name = [var_name]
            assert len(var_name) == self.dim
            self.var_name = var_name

    @property
    def random_seed(self):
        return self._random_seed
    
    @random_seed.setter
    def random_seed(self, seed):
        if seed:
            self._random_seed = int(seed)
            np.random.seed(self._random_seed)

    @abstractmethod
    def sampling(self, N=1):
        """The output is a list of shape (N, self.dim)
        """
        pass
    
    def _set_index(self):
        self.C_mask = np.asarray(self.var_type) == 'C'  # Continuous
        self.O_mask = np.asarray(self.var_type) == 'O'  # Ordinal
        self.N_mask = np.asarray(self.var_type) == 'N'  # Nominal 
        
        self.id_C = np.nonzero(self.C_mask)[0]
        self.id_O = np.nonzero(self.O_mask)[0]
        self.id_N = np.nonzero(self.N_mask)[0]

    def _set_levels(self):
        """Set categorical levels for all nominal variables
        """
        if hasattr(self, 'id_N') and len(self.id_N) > 0:
            self.levels = {i : self.bounds[i] for i in self.id_N}
            self._n_levels = {i : len(self.bounds[i]) for i in self.id_N}
        else:
            self.levels, self._n_levels = None, None 

    def to_linear_scale(self, X):
        X = deepcopy(X)
        if not hasattr(X[0], '__iter__'):
            for k, v in self.scale.items():
                X[k] = INV_TRANS[v](X[k])
        else:
            for k, v in self.scale.items():
                for i in range(len(X)):
                    X[i][k] = INV_TRANS[v](X[i][k])
        return X

    def round(self, X):
        """Round the real-valued components of `X` to the 
        corresponding numerical precision, if given
        """
        # NOTE: make sure the rounding is applied in the original linear scale
        X = self.to_linear_scale(X)

        if self.precision is not None:
            X = deepcopy(X)
            if not hasattr(X[0], '__iter__'):
                for k, v in self.precision.items():
                    X[k] = np.round(X[k], v)
            else:
                for k, v in self.precision.items():
                    for i in range(len(X)):
                        X[i][k] = np.round(X[i][k], v)
        return X

    @classmethod
    def from_dict(cls, param, space_name=True):
        """Create a search space object from input dictionary

        Parameters
        ----------
        param : dict
            A dictionary that describes the search space
        space_name : bool, optional
            Whether a (multi-dimensional) subspace should be named. If this named space 
            is a subspace a whole search space, for a solution sampled from the whole space, its 
            components pertaining to this subspace will be grouped together under the key 
            `space_name`, when this solution is converted to a dictionary/json
            (see `SearchSpace.to_dict`).

        Returns
        -------
        SearchSpace
        """
        assert isinstance(param, dict)

        # construct the search space
        for i, (k, v) in enumerate(param.items()):
            bounds = v['range']
            if not hasattr(bounds[0], '__iter__') or isinstance(bounds[0], str):
                bounds = [bounds]

            N = v['N'] if 'N' in v else int(1)
            bounds *= N
            name = k if space_name else None

            # IMPORTANT: name argument is necessary for the variable grouping
            if v['type'] in ['r', 'real']:                  # real-valued parameter
                precision = v['precision'] if 'precision' in v else None 
                scale = v['scale'] if 'scale' in v else None 
                space_ = ContinuousSpace(
                    bounds, var_name=k, name=name, 
                    precision=precision, scale=scale
                )
            elif v['type'] in ['i', 'int', 'integer']:      # integer-valued parameter
                space_ = OrdinalSpace(bounds, var_name=k, name=name)
            elif v['type'] in ['c', 'cat', 'bool']:         # category-valued parameter
                space_ = NominalSpace(bounds, var_name=k, name=name) 
            
            if i == 0:
                space = space_
            else:
                space += space_
        return space

    @classmethod
    def from_json(cls, file):
        """Create a seach space from a json file

        Parameters
        ----------
        file : str
            Path to the input json file

        Returns
        -------
        SearchSpace
            an `SearchSpace` object converted from the json file
        """
        with open(file, 'r') as f:
            return cls.from_dict(json.load(f))

    def __len__(self):
        return self.dim

    def __iter__(self):
        pass

    def __add__(self, space):
        """Direct Sum of two `SearchSpace`s
        """
        assert isinstance(space, SearchSpace)
        return ProductSpace(self, space)

    def __radd__(self, space):
        return self.__add__(space)

    def __mul__(self, N):
        """Replicate a `SearchSpace` N times
        """
        N = int(N)
        s = deepcopy(self)
        s.dim = int(self.dim * N)
        s.var_type *= N
        s.bounds *= N
        s.var_name = ['{}_{}'.format(v, k) for k in range(N) for v in self.var_name]
        return s

    def __rmul__(self, N):
        return self.__mul__(N)
    
    def __repr__(self):
        return self.__str__()

    def __str__(self):
        _ = 'Search Space of %d variables: \n'%self.dim
        for i in range(self.dim):
            _ += '   `%s`'%self.var_name[i]
            _ += ' - categories: ' if self.var_type[i] == 'N' else ' bounds: ' 
            _ += str(self.bounds[i])
            if i in self.precision:
                _ += ' - precision: %d'%self.precision[i]
            if i in self.scale:
                _ += ' - scale: %s'%self.scale[i]
            _ += '\n'
        return _
        

class ContinuousSpace(SearchSpace):
    """Continuous (real-valued) Search Space
    """
    def __init__(
        self, 
        bounds, 
        var_name='r', 
        name=None, 
        precision=None,
        scale=None
        ):
        super(ContinuousSpace, self).__init__(bounds, var_name, name)
        self.var_type = ['C'] * self.dim
        self._set_index()
        
        # set up the precision for each dimension
        if hasattr(precision, '__iter__'):
            assert len(precision) == self.dim
            self.precision = {
                i : precision[i] for i in range(self.dim) if precision[i] is not None
            }
        elif precision is not None:
            self.precision = {i : precision for i in range(self.dim)}

        # set up the scale for each dimension
        if scale is not None:
            if isinstance(scale, str):
                scale = [scale] * self.dim
            elif hasattr(scale, '__iter__'):
                assert len(scale) == self.dim

            self.scale = {
                i : scale[i] for i in range(self.dim) if scale[i] is not None
            }

        for i, s in self.scale.items():
            lower, upper = self.bounds[i]
            self.bounds[i] = (TRANS[s](lower), TRANS[s](upper))

        self._bounds = np.atleast_2d(self.bounds).T
        assert all(self._bounds[0, :] < self._bounds[1, :])

    def __mul__(self, N):
        s = super(ContinuousSpace, self).__mul__(N)
        s._bounds = np.tile(s._bounds, (1, N))
        s._set_index()

        s.precision = {}
        for i in range(N):
            s.precision.update(
                {(k + self.dim * i) : v for k, v in self.precision.items()}
            )

        s.scale = {}
        for i in range(N):
            s.scale.update(
                {(k + self.dim * i) : v for k, v in self.scale.items()}
            )
        return s
    
    def sampling(self, N=1, method='uniform'):
        lb, ub = self._bounds
        if method == 'uniform':   # uniform random samples
            X = ((ub - lb) * rand(N, self.dim) + lb)
        elif method == 'LHS':     # Latin hypercube sampling
            if N == 1:
                X = ((ub - lb) * rand(N, self.dim) + lb)
            else:
                X = ((ub - lb) * lhs(self.dim, samples=N, criterion='maximin') + lb)
        return X.tolist()


class NominalSpace(SearchSpace):
    """Nominal (discrete) Search Space
    """
    def __init__(self, levels, var_name='d', name=None):
        levels = self._get_unique_levels(levels)
        super(NominalSpace, self).__init__(levels, var_name, name)
        self.var_type = ['N'] * self.dim
        self._levels = [np.array(b) for b in self.bounds]
        self._set_index()
        self._set_levels()
    
    def _get_unique_levels(self, levels):
        index = list(hasattr(l, '__iter__') and not isinstance(l, str) for l in levels)
        if any(index):
            return [
                list(set(levels[k] if i else [levels[k]])) \
                    for k, i in enumerate(index)
            ]
        else: 
            return [list(set(levels))]

    def __mul__(self, N):
        s = super(NominalSpace, self).__mul__(N)
        s._set_index()
        s._set_levels()
        return s
    
    def sampling(self, N=1, method='uniform'):
        # NOTE: `LHS` sampling does not apply here since nominal variable is not ordered
        res = np.empty((N, self.dim), dtype=object)
        for i in range(self.dim):
            idx = randint(0, self._n_levels[i], N)
            res[:, i] = [self.levels[i][_] for _ in idx]

        return res.tolist()


class OrdinalSpace(SearchSpace):
    """Ordinal (integer) Search Space
    """
    def __init__(self, bounds, var_name='i', name=None):
        super(OrdinalSpace, self).__init__(bounds, var_name, name)
        self.var_type = ['O'] * self.dim
        self._lb, self._ub = zip(*self.bounds)        # for sampling
        assert all(np.array(self._lb) < np.array(self._ub))
        self._set_index()

    def __mul__(self, N):
        s = super(OrdinalSpace, self).__mul__(N)
        s._lb, s._ub = s._lb * N, s._ub * N
        s._set_index()
        return s
    
    def sampling(self, N=1, method='uniform'):
        # TODO: adding LHS sampling here
        res = np.zeros((N, self.dim), dtype=int)
        for i in range(self.dim):
            res[:, i] = list(map(int, randint(self._lb[i], self._ub[i], N)))
        return res.tolist()


class ProductSpace(SearchSpace):
    """Cartesian product of the search spaces
    """
    def __init__(self, spaceL, spaceR):
        # setup the space names
        nameL = spaceL.name if isinstance(spaceL, ProductSpace) else [spaceL.name] 
        nameR = spaceR.name if isinstance(spaceR, ProductSpace) else [spaceR.name]
        self.name = nameL + nameR
        self.dim = spaceL.dim + spaceR.dim

        # TODO: check coincides of variable names
        self.var_name = spaceL.var_name + spaceR.var_name             
        self.bounds = spaceL.bounds + spaceR.bounds
        self.var_type = spaceL.var_type + spaceR.var_type

        self._subspaceL = deepcopy(spaceL)
        self._subspaceR = deepcopy(spaceR)
        self._set_index()
        self._set_levels()

        self.precision = copy(spaceL.precision)
        self.precision.update({(k + spaceL.dim) : v for k, v in spaceR.precision.items()})

        self.scale = copy(spaceL.scale) 
        self.scale.update({(k + spaceL.dim) : v for k, v in spaceR.scale.items()})
    
    def sampling(self, N=1, method='uniform'):
        a = self._subspaceL.sampling(N, method)
        b = self._subspaceR.sampling(N, method)
        return [a[i] + b[i] for i in range(N)]
    
    def to_dict(self, solution):
        """Save a Solution instance to a dictionary 
        
        The result is grouped by sub-spaces, which is meant for vector-valued 
        parameters for the configuration 

        Parameters
        ----------
        solution : .base.Solution
            A solution object

        Returns
        -------
        dict
        """
        id1 = list(range(self._subspaceL.dim))
        id2 = list(range(self._subspaceL.dim, self.dim))
        L = solution[id1] if len(solution.shape) == 1 else solution[:, id1]
        R = solution[id2] if len(solution.shape) == 1 else solution[:, id2]
        return {**self._subspaceL.to_dict(L), **self._subspaceR.to_dict(R)}

    def __mul__(self, space):
        raise ValueError('Unsupported operation')

    def __rmul__(self, space):
        raise ValueError('Unsupported operation')



# -*- coding: utf-8 -*-
"""
Created on Mon May 19 10:17:43 2014

@author: wangronin
"""

import numpy as np
from numpy import isfinite, mod, floor, shape, bitwise_and, zeros, newaxis

def boundary_handling(x, lb, ub):
    """
    
    This function transforms x to t w.r.t. the low and high
    boundaries lb and ub. It implements the function T^{r}_{[a,b]} as
    described in Rui Li's PhD thesis "Mixed-Integer Evolution Strategies
    for Parameter Optimization and Their Applications to Medical Image 
    Analysis" as alorithm 6.
    
    """
    
    lb, ub = lb.flatten(), ub.flatten()
    
    lb_index = isfinite(lb)
    up_index = isfinite(ub)
    
    valid = bitwise_and(lb_index,  up_index)
    
    LB = lb[valid][:, newaxis]
    UB = ub[valid][:, newaxis]

    y = (x[valid, :] - LB) / (UB - LB)
    I = mod(floor(y), 2) == 0
    yprime = zeros(shape(y))
    yprime[I] = np.abs(y[I] - floor(y[I]))
    yprime[~I] = 1.0 - np.abs(y[~I] - floor(y[~I]))

    x[valid, :] = LB + (UB - LB) * yprime
        
    return x


"""
Created on Wed Aug 23 15:48:28 2017

@author: Hao Wang
@email: wangronin@gmail.com
Note: this module is strongly inspired by the kernel module of the sklearn GaussianProcess kernel implementation.
"""


from abc import abstractmethod
import numpy as np
import math
from scipy.special import kv, gamma

"""
The built-in correlation models submodule for the gaussian_process module.
TODO: The grad of the correlation should be implemented in the 
correlation models

TODO: by default, all the stationary kernels are normalized
"""

class Kernel(object):
    def __init__(self):
        pass

    @abstractmethod
    def __call__(self, X, Y=None):
        "Evaluate the kernel function"
    
    def __add__(self, kernel):
        return KernelSum(self, kernel)

    def __mul__(self, kernel):
        return KernelProduct(self, kernel)

    def __rmul__(self, kernel):
        return KernelProduct(self, kernel)

class ConstantKernel(Kernel):
    def __init__(self, sigma2=1.0):
        self.sigma2 = sigma2
    
    def __call__(self, X, Y=None):
        pass

class CompositeKernel(Kernel):
    """
    The space of kernels is closed under addition and multiplication 
    """
    def __init__(self):
        pass

class KernelSum(CompositeKernel):
    def __call__(self, X, Y=None):
        return self.K1(X, Y) + self.K2(X, Y)

class KernelProduct(CompositeKernel):
    def __call__(self, X, Y=None):
        return self.K1(X, Y) * self.K2(X, Y)

class HammingKernel(Kernel):
    """
    Kernel function for categorical variables using Hamming distance
    """
    def __call__(self, X, Y=None):
        if Y is None:
            Y = X
        
class StationaryKernel(Kernel):
    pass

class Matern(StationaryKernel):
    def __init__(self, theta=None, bounds=None, nu=1.5):
        self.nu = nu
        self.theta = theta
        self.bounds = self.theta_bounds() if bounds else bounds
    
    def __call__(self, X):

        theta = np.asarray(theta, dtype=np.float64)
        X = np.asarray(X, dtype=np.float64)
        if X.ndim > 1:
            n_features = X.shape[1]
        else:
            n_features = 1

        if theta.size == 1:
            dists = np.sqrt(theta[0] * np.sum(X ** 2, axis=1))
        else:
            dists = np.sqrt(np.sum(theta.reshape(1, n_features) * X ** 2, axis=1))
         # Matern 1/2
        if nu == 0.5:
            K = np.exp(-dists)
        # Matern 3/2
        elif nu == 1.5:

            K = dists * math.sqrt(3)
            K = (1. + K) * np.exp(-K)
        # Matern 5/2
        elif nu == 2.5:
            K = dists * math.sqrt(5)
            K = (1. + K + K ** 2 / 3.0) * np.exp(-K)
        else:  # general case; expensive to evaluate
            K = dists
            K[K == 0.0] += np.finfo(float).eps  # strict zeros result in nan
            tmp = (math.sqrt(2 * nu) * K)
            K.fill((2 ** (1. - nu)) / gamma(nu))
            K *= tmp ** nu
            K *= kv(nu, tmp)
    
    def dx(self, X):
       pass
    
    def dtheta(self, X):
        c = np.sqrt(3)
        D = np.sqrt(np.sum(theta * diff, axis=-1))

        if nu == 0.5:
            grad = - diff * theta / D * R
        elif nu == 1.5:
            grad = -3 * np.exp(-c * D)[..., np.newaxis] * diff / 2.
        elif nu == 2.5:
            pass

# Note: the c-wrapper for kernel functions
#def matern(theta, X, eval_Dx=False, eval_Dtheta=False,
#           length_scale_bounds=(1e-5, 1e5), nu=1.5):
#    n_max = dim*(dim-1)/2
#    
#    index = (rand(n_max) < pm) * arange(1, n_max+1)
#    index = index[index > 0]
#    index = index - 1
#    
#    n_rotation = index.shape[0]
#    rotations = angle * (2*rand(n_rotation) - 1)
##        res = zeros((dim, dim))
#    
#    # Prepare the arguments for C function
#    POS = index.ctypes.data_as(POINTER(c_uint))
#    ROTATIONS = rotations.ctypes.data_as(POINTER(c_double))
#    RES = operator.ctypes.data_as(POINTER(c_double))
#    
#    # Invoke the C code for dense matrix multiplication
#    c_noisy_gen(dim, POS, n_rotation, ROTATIONS, RES)

def matern(theta, X, eval_Dx=False, eval_Dtheta=False,
           length_scale_bounds=(1e-5, 1e5), nu=1.5):
    """
    theta = np.asarray(theta, dtype=np.float64)
    d = np.asarray(d, dtype=np.float64)

    if d.ndim > 1:
        n_features = d.shape[1]
    else:
        n_features = 1

    if theta.size == 1:
        return np.exp(-theta[0] * np.sum(d ** 2, axis=1))
    elif theta.size != n_features:
        raise ValueError("Length of theta must be 1 or %s" % n_features)
    else:
        return np.exp(-np.sum(theta.reshape(1, n_features) * d ** 2, axis=1))

    """
    theta = np.asarray(theta, dtype=np.float64)
    X = np.asarray(X, dtype=np.float64)
    if X.ndim > 1:
        n_features = X.shape[1]
    else:
        n_features = 1

    if theta.size == 1:
        dists = np.sqrt(theta[0] * np.sum(X ** 2, axis=1))
    else:
        dists = np.sqrt(np.sum(theta.reshape(1, n_features) * X ** 2, axis=1))

    # Matern 1/2
    if nu == 0.5:
        K = np.exp(-dists)
    # Matern 3/2
    elif nu == 1.5:

        K = dists * math.sqrt(3)
        K = (1. + K) * np.exp(-K)
    # Matern 5/2
    elif nu == 2.5:
        K = dists * math.sqrt(5)
        K = (1. + K + K ** 2 / 3.0) * np.exp(-K)
    else:  # general case; expensive to evaluate
        K = dists
        K[K == 0.0] += np.finfo(float).eps  # strict zeros result in nan
        tmp = (math.sqrt(2 * nu) * K)
        K.fill((2 ** (1. - nu)) / gamma(nu))
        K *= tmp ** nu
        K *= kv(nu, tmp)

    if eval_Dx:
        pass

    # infoert from upper-triangular matrix to square matrix
    # K = squareform(K)
    # np.fill_diagonal(K, 1)
    if eval_Dtheta:
        pass
#        # We need to recompute the pairwise dimension-wise distances
#        if self.anisotropic:
#            D = (X[:, np.newaxis, :] - X[np.newaxis, :, :])**2 \
#                / (length_scale ** 2)
#        else:
#            D = squareform(dists**2)[:, :, np.newaxis]
#
#        if self.nu == 0.5:
#            K_gradient = K[..., np.newaxis] * D \
#                / np.sqrt(D.sum(2))[:, :, np.newaxis]
#            K_gradient[~np.isfinite(K_gradient)] = 0
#        elif self.nu == 1.5:
#            K_gradient = \
#                3 * D * np.exp(-np.sqrt(3 * D.sum(-1)))[..., np.newaxis]
#        elif self.nu == 2.5:
#            tmp = np.sqrt(5 * D.sum(-1))[..., np.newaxis]
#            K_gradient = 5.0 / 3.0 * D * (tmp + 1) * np.exp(-tmp)
#        else:
#            # approximate gradient numerically
#            def f(theta):  # helper function
#                return self.clone_with_theta(theta)(X, Y)
#            return K, _approx_fprime(self.theta, f, 1e-10)
#
#        if not self.anisotropic:
#            return K, K_gradient[:, :].sum(-1)[:, :, np.newaxis]
#        else:
#            return K, K_gradient
    return K

def absolute_exponential(theta, d):
    """
    Absolute exponential autocorrelation model.
    (Ornstein-Uhlenbeck stochastic process)::

                                          n
        theta, d --> r(theta, d) = exp(  sum  - theta_i * |d_i| )
                                        i = 1

    Parameters
    ----------
    theta : array_like
        An array with shape 1 (isotropic) or n (anisotropic) giving the
        autocorrelation parameter(s).

    d : array_like
        An array with shape (n_eval, n_features) giving the componentwise
        distances between locations x and x' at which the correlation model
        should be evaluated.

    Returns
    -------
    r : array_like
        An array with shape (n_eval, ) containing the values of the
        autocorrelation model.
    """
    theta = np.asarray(theta, dtype=np.float64)
    d = np.abs(np.asarray(d, dtype=np.float64))

    if d.ndim > 1:
        n_features = d.shape[1]
    else:
        n_features = 1

    if theta.size == 1:
        return np.exp(- theta[0] * np.sum(d, axis=1))
    elif theta.size != n_features:
        raise ValueError("Length of theta must be 1 or %s" % n_features)
    else:
        return np.exp(- np.sum(theta.reshape(1, n_features) * d, axis=1))


def squared_exponential(theta, d):
    """
    Squared exponential correlation model (Radial Basis Function).
    (Infinitely differentiable stochastic process, very smooth)::

                                          n
        theta, d --> r(theta, d) = exp(  sum  - theta_i * (d_i)^2 )
                                        i = 1

    Parameters
    ----------
    theta : array_like
        An array with shape 1 (isotropic) or n (anisotropic) giving the
        autocorrelation parameter(s).

    d : array_like
        An array with shape (n_eval, n_features) giving the componentwise
        distances between locations x and x' at which the correlation model
        should be evaluated.

    Returns
    -------
    r : array_like
        An array with shape (n_eval, ) containing the values of the
        autocorrelation model.
    """

    theta = np.asarray(theta, dtype=np.float64)
    d = np.asarray(d, dtype=np.float64)

    if d.ndim > 1:
        n_features = d.shape[1]
    else:
        n_features = 1

    if theta.size == 1:
        return np.exp(-theta[0] * np.sum(d ** 2, axis=1))
    elif theta.size != n_features:
        raise ValueError("Length of theta must be 1 or %s" % n_features)
    else:
        return np.exp(-np.sum(theta.reshape(1, n_features) * d ** 2, axis=1))


def generalized_exponential(theta, d):
    """
    Generalized exponential correlation model.
    (Useful when one does not know the smoothness of the function to be
    predicted.)::

                                          n
        theta, d --> r(theta, d) = exp(  sum  - theta_i * |d_i|^p )
                                        i = 1

    Parameters
    ----------
    theta : array_like
        An array with shape 1+1 (isotropic) or n+1 (anisotropic) giving the
        autocorrelation parameter(s) (theta, p).

    d : array_like
        An array with shape (n_eval, n_features) giving the componentwise
        distances between locations x and x' at which the correlation model
        should be evaluated.

    Returns
    -------
    r : array_like
        An array with shape (n_eval, ) with the values of the autocorrelation
        model.
    """

    theta = np.asarray(theta, dtype=np.float64)
    d = np.asarray(d, dtype=np.float64)

    if d.ndim > 1:
        n_features = d.shape[1]
    else:
        n_features = 1

    lth = theta.size
    if n_features > 1 and lth == 2:
        theta = np.hstack([np.repeat(theta[0], n_features), theta[1]])
    elif lth != n_features + 1:
        raise Exception("Length of theta must be 2 or %s" % (n_features + 1))
    else:
        theta = theta.reshape(1, lth)

    td = theta[:, 0:-1].reshape(1, n_features) * np.abs(d) ** theta[:, -1]
    r = np.exp(- np.sum(td, 1))

    return r


def pure_nugget(theta, d):
    """
    Spatial independence correlation model (pure nugget).
    (Useful when one wants to solve an ordinary least squares problem!)::

                                           n
        theta, d --> r(theta, d) = 1 if   sum |d_i| == 0
                                         i = 1
                                   0 otherwise

    Parameters
    ----------
    theta : array_like
        None.

    d : array_like
        An array with shape (n_eval, n_features) giving the componentwise
        distances between locations x and x' at which the correlation model
        should be evaluated.

    Returns
    -------
    r : array_like
        An array with shape (n_eval, ) with the values of the autocorrelation
        model.
    """

    theta = np.asarray(theta, dtype=np.float64)
    d = np.asarray(d, dtype=np.float64)

    n_eval = d.shape[0]
    r = np.zeros(n_eval)
    r[np.all(d == 0., axis=1)] = 1.

    return r


def cubic(theta, d):
    """
    Cubic correlation model::

        theta, d --> r(theta, d) =
          n
         prod max(0, 1 - 3(theta_j*d_ij)^2 + 2(theta_j*d_ij)^3) ,  i = 1,...,m
        j = 1

    Parameters
    ----------
    theta : array_like
        An array with shape 1 (isotropic) or n (anisotropic) giving the
        autocorrelation parameter(s).

    d : array_like
        An array with shape (n_eval, n_features) giving the componentwise
        distances between locations x and x' at which the correlation model
        should be evaluated.

    Returns
    -------
    r : array_like
        An array with shape (n_eval, ) with the values of the autocorrelation
        model.
    """

    theta = np.asarray(theta, dtype=np.float64)
    d = np.asarray(d, dtype=np.float64)

    if d.ndim > 1:
        n_features = d.shape[1]
    else:
        n_features = 1

    lth = theta.size
    if lth == 1:
        td = np.abs(d) * theta
    elif lth != n_features:
        raise Exception("Length of theta must be 1 or " + str(n_features))
    else:
        td = np.abs(d) * theta.reshape(1, n_features)

    td[td > 1.] = 1.
    ss = 1. - td ** 2. * (3. - 2. * td)
    r = np.prod(ss, 1)

    return r



"""
Created on Mon Sep 11 10:48:14 2017

@author: Hao Wang
@email: wangronin@gmail.com
"""
import numpy as np
from numpy import std, array, atleast_2d

from sklearn.ensemble import RandomForestRegressor
from sklearn.utils.validation import check_is_fitted
from sklearn.ensemble._base import _partition_estimators
from sklearn.preprocessing import OneHotEncoder, LabelEncoder
from sklearn.metrics import r2_score

from joblib import Parallel, delayed

# TODO: implement multi-output/objetive surrogate models, better to model the c
# orrelation among targets
class SurrogateAggregation(object):
    def __init__(self, surrogates, aggregation='WS', **kwargs):
        self.surrogates = surrogates
        self.N = len(self.surrogates)
        self.aggregation = aggregation
        self.weights = np.asarray(kwargs['weights'], dtype='float').ravel()

        assert self.aggregation in ['WS', 'Tchebycheff']

    def fit(self, X, y):
        pass

    def predict(self, X, eval_MSE=False):
        if eval_MSE:
            y_hat_, MSE_ = list(zip(*[s.predict(X, eval_MSE=True) for s in self.surrogates]))
            y_hat_ = np.atleast_2d([_.ravel() for _ in y_hat_])
            MSE_ = np.atleast_2d([_.ravel() for _ in MSE_])
        else:
            y_hat_ = np.atleast_2d([_.predict(X, eval_MSE=False).ravel() for _ in self.surrogates])

        if self.aggregation == 'WS':
            y_hat = self.weights.dot(y_hat_)
            if eval_MSE:
                MSE = (self.weights ** 2.).dot(MSE_)

        elif self.aggregation == 'Tchebycheff':
            # TODO: implement this part
            pass
        
        return (y_hat, MSE) if eval_MSE else y_hat
    
    def gradient(self, X):
        # TODO: implement
        pass
        
def _save_prediction(predict, X, index, out):
    """
    It can't go locally in ForestClassifier or ForestRegressor, because joblib
    complains that it cannot pickle it when placed there.
    """
    out[..., index] = predict(X, check_input=False)

class RandomForest(RandomForestRegressor):
    """Extension on the sklearn's `RandomForestRegressor`
    Added functionality: 
        1) MSE estimate, 
        2) OneHotEncoding to handle categorical variables
    """
    def __init__(
        self, 
        n_estimators=100, 
        max_features=5/6, 
        min_samples_leaf=2, 
        levels=None, 
        **kwargs
        ):
        """
        parameter
        ---------
        levels : dict, for categorical inputs
            keys: indices of categorical variables
            values: list of levels of categorical variables
        """
        super(RandomForest, self).__init__(
            n_estimators=n_estimators,
            max_features=max_features,
            min_samples_leaf=min_samples_leaf,
            **kwargs
        )
        self.is_fitted = False

        # for categorical levels/variable number
        # in the future, maybe implement binary/multi-value split
        if levels is not None:
            assert isinstance(levels, dict)
            self._levels = levels
            self._cat_idx = list(self._levels.keys())
            self._categories = [list(l) for l in self._levels.values()]

            # encode categorical variables to binary values
            self._enc = OneHotEncoder(categories=self._categories, sparse=False)

    def _check_X(self, X):
        X_ = array(X, dtype=object)
        if hasattr(self, '_levels'):
            X_cat = X_[:, self._cat_idx]
            try:
                X_cat = self._enc.transform(X_cat)
            except:
                X_cat = self._enc.fit_transform(X_cat)
            X = np.c_[np.delete(X_, self._cat_idx, 1).astype(float), X_cat]
        return X

    def fit(self, X, y):
        X = self._check_X(X)
        y = y.ravel()
        self.y = y
        self.is_fitted = True
        return super(RandomForest, self).fit(X, y)

    def predict(self, X, eval_MSE=False):
        """Predict regression target for `X`.
        The predicted regression target of an input sample is computed as the
        mean predicted regression targets of the trees in the forest.
        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, its dtype will be converted to
            ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csr_matrix``.
        Returns
        -------
        y : ndarray of shape (n_samples,) or (n_samples, n_outputs)
            The predicted values.
        """
        check_is_fitted(self)
        # Check data
        X = self._check_X(X)
        X = self._validate_X_predict(X)

        # Assign chunk of trees to jobs
        n_jobs, _, _ = _partition_estimators(self.n_estimators, self.n_jobs)

        # storing the output of every estimator since those are required to estimate the MSE
        if self.n_outputs_ > 1:
            y_hat_all = np.zeros(
                (X.shape[0], self.n_outputs_, self.n_estimators), dtype=np.float64
            )
        else:
            y_hat_all = np.zeros((X.shape[0], self.n_estimators), dtype=np.float64)

        # Parallel loop
        Parallel(n_jobs=n_jobs, verbose=self.verbose, backend="threading")(
            delayed(_save_prediction)(e.predict, X, i, y_hat_all) \
                for i, e in enumerate(self.estimators_)
        )

        y_hat = np.mean(y_hat_all, axis=1).flatten()
        if eval_MSE:
            # TODO: implement the jackknife estimate of variance
            _MSE_hat = np.std(y_hat_all, axis=1, ddof=1) ** 2.
            _MSE_hat = _MSE_hat.flatten()

        return (y_hat, _MSE_hat) if eval_MSE else y_hat

if __name__ == '__main__':
    # TODO: this part goes into test 
    # simple test for mixed variables...
    np.random.seed(12)

    n_sample = 110
    levels = ['OK', 'A', 'B', 'C', 'D', 'E']
    X = np.c_[np.random.randn(n_sample, 2).astype(object),
              np.random.choice(levels, size=(n_sample, 1))]
    y = np.sum(X[:, 0:-1] ** 2., axis=1) + 5 * (X[:, -1] == 'OK')

    X_train, y_train = X[:100, :], y[:100]
    X_test, y_test = X[100:, :], y[100:]

    # sklearn-random forest
    rf = RandomForest(levels={2: levels}, max_features='sqrt')
    rf.fit(X_train, y_train)
    y_hat, mse = rf.predict(X_test, eval_MSE=True)

    print('sklearn random forest:')
    print('target :', y_test)
    print('predicted:', y_hat)
    print('MSE:', mse)
    print('r2:', r2_score(y_test, y_hat))
    print()

    if 11 < 2:
        X = np.c_[np.random.randn(n_sample, 2).astype(object),
                  np.random.choice(levels, size=(n_sample, 1))]
        y = np.sum(X[:, 0:-1] ** 2., axis=1) + 5 * (X[:, -1] == 'OK')

        X_train, y_train = X[:100, :], y[:100]
        X_test, y_test = X[100:, :], y[100:]

        rf_ = RandomForest(levels={2: levels}, max_features='sqrt')
        rf_.fit(X_train, y_train)

        rf_aggr = SurrogateAggregation((rf, rf_), weights=(0.1, 0.9))
        y_hat, mse = rf_aggr.predict(X_test, eval_MSE=True)

        print('sklearn random forest:')
        print('target :', y_test)
        print('predicted:', y_hat)
        print('MSE:', mse)
        print('r2:', r2_score(y_test, y_hat))
        print()

    # TODO: those settings should be in test file as inputs to surroagtes
    # leaf_size = max(1, int(n_sample / 20.))
    # ntree=100,
    # mtry=ceil(self.n_feature * 5 / 6.),
    # nodesize=leaf_size



"""

@author: Hao Wang
@email: wangronin@gmail.com

"""
from typing import Callable, Any, Tuple
import os, sys, dill, functools, logging, time

import numpy as np
import json, copy, re 
from copy import copy
from joblib import Parallel, delayed

from . import AcquisitionFunction
from .base import baseBO
from .Solution import Solution
from .SearchSpace import SearchSpace

class BO(baseBO):
    def _create_acquisition(self, fun=None, par={}, return_dx=False):
        fun = fun if fun is not None else self._acquisition_fun
        if hasattr(getattr(AcquisitionFunction, fun), 'plugin'):
            if 'plugin' not in par:
                par.update({'plugin': self.fmin})
        
        return super()._create_acquisition(fun, par, return_dx)

    def pre_eval_check(self, X):
        """Check for the duplicated solutions as it is not allowed in noiseless cases
        """
        if not isinstance(X, Solution):
            X = Solution(X, var_name=self.var_names)
        
        N = X.N
        if hasattr(self, 'data'):
            X = X + self.data

        _ = []
        for i in range(N):
            x = X[i]
            idx = np.arange(len(X)) != i
            CON = np.all(
                np.isclose(
                    np.asarray(X[idx][:, self.r_index], dtype='float'),
                    np.asarray(x[self.r_index], dtype='float')
                ), axis=1
            )
            INT = np.all(X[idx][:, self.i_index] == x[self.i_index], axis=1)
            CAT = np.all(X[idx][:, self.d_index] == x[self.d_index], axis=1)
            if not any(CON & INT & CAT):
                _ += [i]

        return X[_]

class ParallelBO(BO):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        assert self.n_point > 1

        if self._acquisition_fun == 'MGFI':
            self._par_name = 't'
            # Log-normal distribution for `t` supported on [0, \infty)
            self._sampler = lambda x: np.exp(np.log(x['t']) + 0.5 * np.random.randn())
        elif self._acquisition_fun == 'UCB':
            self._par_name = 'alpha'
            # Logit-normal distribution for `alpha` supported on [0, 1]
            self._sampler = lambda x: 1 / (1 + np.exp((x['alpha'] * 4 - 2) \
                + 0.6 * np.random.randn())) 
        elif self._acquisition_fun == 'EpsilonPI':
            self._par_name = 'epsilon'
            self._sampler = None # TODO: implement this!
        else:
            raise NotImplementedError
        
        _criterion = getattr(AcquisitionFunction, self._acquisition_fun)()
        if self._par_name not in self._acquisition_par:
            self._acquisition_par[self._par_name] = getattr(_criterion, self._par_name)
        
    def _batch_arg_max_acquisition(self, n_point, return_dx):
        criteria = []
        for _ in range(n_point):
            _par = self._sampler(self._acquisition_par)
            _acquisition_par = copy(self._acquisition_par)
            _acquisition_par.update({self._par_name : _par})
            criteria.append(
                self._create_acquisition(par=_acquisition_par, return_dx=return_dx)
            )
        
        if self.n_job > 1:
            __ = Parallel(n_jobs=self.n_job)(
                delayed(self._argmax_restart)(c, logger=self._logger) for c in criteria
            )
        else:
            __ = [list(self._argmax_restart(_, logger=self._logger)) for _ in criteria]
        
        return tuple(zip(*__))

class AnnealingBO(ParallelBO):
    def __init__(self, t0=2, tf=1e-1, schedule='exp', *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        self.t0 = t0
        self.tf = tf
        self.schedule = schedule
        self._acquisition_par['t'] = t0
        
        # TODO: add supports for UCB and epsilon-PI
        max_iter = (self.max_FEs - self._DoE_size) / self.n_point
        if self.schedule == 'exp':                          # exponential
            self.alpha = (self.tf / t0) ** (1. / max_iter) 
            self._anealer = lambda t: t * self.alpha
        elif self.schedule == 'linear':
            self.eta = (t0 - self.tf) / max_iter            # linear
            self._anealer = lambda t: t - self.eta
        elif self.schedule == 'log':
            self.c = self.tf * np.log(max_iter + 1)         # logarithmic 
            self._anealer = lambda t: t * self.c / np.log(self.iter_count + 2)
        else:
            raise NotImplementedError
        
        def callback():
            self._acquisition_par['t'] = self._anealer(self._acquisition_par['t'])
        self._acquisition_callbacks += [callback]
    
class SelfAdaptiveBO(ParallelBO):
    def __init__(self, *argv, **kwargs):
        super.__init__(*argv, **kwargs)
        assert self.n_point > 1

    def _batch_arg_max_acquisition(self, n_point, return_dx):
        criteria = []
        _t_list = []
        N = int(n_point / 2)

        for _ in range(n_point):
            _t = np.exp(self._acquisition_par['t'] * np.random.randn())
            _t_list.append(_t)
            _acquisition_par = copy(self._acquisition_par)
            _acquisition_par.update({'t' : _t})
            criteria.append(
                self._create_acquisition(par=_acquisition_par, return_dx=return_dx)
            )
        
        if self.n_job > 1:
            __ = Parallel(n_jobs=self.n_job)(
                delayed(self._argmax_restart)(c, logger=self._logger) for c in criteria
            )
        else:
            __ = [list(self._argmax_restart(_, logger=self._logger)) for _ in criteria]
        
        # NOTE: this adaptation is different from what I did in the LeGO paper..
        idx = np.argsort(__[1])[::-1][:N]
        self._acquisition_par['t'] = np.mean([_t_list[i] for i in idx])
        return tuple(zip(*__))

class NoisyBO(ParallelBO):
    def pre_eval_check(self, X):
        if not isinstance(X, Solution):
            X = Solution(X, var_name=self.var_names)
        return X

    def _create_acquisition(self, fun=None, par={}, return_dx=False):
        if hasattr(getattr(AcquisitionFunction, self._acquisition_fun), 'plugin'):
            # use the model prediction to determine the plugin under noisy scenarios
            # TODO: add more options for determining the plugin value
            y_ = self.model.predict(self.data)
            plugin = np.min(y_) if self.minimize else np.max(y_)
            par.update({'plugin' : plugin})
        
        return super()._create_acquisition(par=par, return_dx=return_dx)


import logging, sys, functools
import numpy as np

from typing import Callable, Union
from copy import copy
from joblib import Parallel, delayed
from scipy.stats import rankdata, norm
from sklearn.decomposition import PCA, KernelPCA

from . import AcquisitionFunction
from .base import baseOptimizer
from .SearchSpace import ContinuousSpace
from .BayesOpt import BO, ParallelBO
from .misc import LoggerFormatter

def penalized_acquisition(x, acquisition_func, X_mean, pca, bounds, return_dx):
    x_ = pca.inverse_transform(x) + X_mean
    bounds = np.asarray(bounds)
    idx_lower = x_ < bounds[:, 0]
    idx_upper = x_ > bounds[:, 1]
    penalty = np.sum([bounds[i, 0] - x_[i] for i in idx_lower]) + \
        np.sum([x_[i] - bounds[i, 1] for i in idx_upper])
    penalty *= -1

    if penalty == 0:
        return acquisition_func(x)
    else:
        if return_dx:
            g = np.zeros((len(x), 1))
            g[idx_lower, :] = 1
            g[idx_upper, :] = -1
            return penalty, g
        else:
            return penalty

class PCABO(ParallelBO):
    def __init__(
        self, 
        kernel_pca: bool = False,
        n_components: Union[float, int] = None,
        **kwargs
        ):
        super().__init__(**kwargs)
        assert isinstance(self._search_space, ContinuousSpace)

        self.__search_space = self._search_space # the original search space
        self.kernel_pca = kernel_pca
        self._n_components = n_components

    def _scale_X(self, X, func_vals):
        self._X_mean = X.mean(axis=0)
        X_ = X - self._X_mean

        if not self.minimize:
            func_vals = -1 * func_vals 

        r = rankdata(func_vals)
        N = len(func_vals)
        w = np.log(N) - np.log(r)
        w /= np.sum(w)
        return X_ * w.reshape(-1, 1)

    def _compute_bounds(self, pca, search_space):
        C = np.array([(l + u) / 2 for l, u in search_space.bounds])
        radius = norm(np.np.array([l for l, _ in search_space.bounds]) - C)
        C = C - pca.mean_ - self._X_mean
        C_ = C.dot(pca.components_.T)
        return [(_ - radius, _ + radius) for _ in C_]

    def _create_acquisition(self, fun=None, par={}, return_dx=False):
        acquisition_func = super()._create_acquisition(fun, par, return_dx)
        fun = functools.partial(
            penalized_acquisition, 
            acquisition_func=acquisition_func, 
            X_mean=self._X_mean, 
            pca=self._pca, 
            bounds=self.__search_space.bounds, 
            return_dx=return_dx
        )
        return fun

    def ask(self, n_point=None):
        X = super().ask(n_point)
        if hasattr(self, '_pca'):
            X = self._pca.inverse_transform(X) + self._X_mean
        return X
    
    def tell(self, X, func_vals):
        X_ = self._scale_X(X, func_vals)

        if self.kernel_pca:
            # TODO: finish the kernel PCA part..
            self._pca = KernelPCA(kernel='rbf', fit_inverse_transform=True, gamma=10)
        else:
            self._pca = PCA(n_components=self._n_components, svd_solver='full')

        X_ = self._pca.fit_transform(X_, func_vals)
        bounds = self._compute_bounds(self._pca, self.__search_space)

        # set the search space in the reduced (feature) space
        self._search_space = ContinuousSpace(bounds)
        super().tell(X_, func_vals)


class OptimizerPipeline(baseOptimizer):
    def __init__(
        self,
        obj_fun: Callable,
        ftarget: float = -np.inf,
        max_FEs: int = None,
        minimize: bool = True,
        verbose: bool = False,
        logger: str = None
        ):
        self.obj_fun = obj_fun
        self.max_FEs = max_FEs
        self.ftarget = ftarget
        self.verbose = verbose
        self.minimize = minimize
        self.queue = []
        self.stop_dict = []
        self.N = 0
        self.logger = logger
        self._counter = 0
        self._curr_opt = None
        self._transfer = None
        self._stop = False

    def add(self, opt, transfer=None):
        opt.obj_fun = self.obj_fun
        opt.verbose = self.verbose
        opt.ftarget = self.ftarget
        opt.minimize = self.minimize
        opt.max_FEs = self.max_FEs
        opt.logger = self._logger

        # add pairs of (optimizer, transfer function)
        self.queue.append((opt, transfer)) 
        self.N += 1

    def ask(self, n_point=None):
        """Get suggestions from the optimizer.

        Parameters
        ----------
        n_point : int
            Desired number of parallel suggestions in the output

        Returns
        -------
        next_guess : list of dict
            List of `n_suggestions` suggestions to evaluate the objective
            function. Each suggestion is a dictionary where each key
            corresponds to a parameter being optimized.
        """
        if not self._curr_opt:
            self._curr_opt, self._transfer = self.queue[self._counter]
            self._logger.name = self._curr_opt.__class__.__name__
        
        return self._curr_opt.ask(n_point=n_point)

    def tell(self, X, y):
        """Feed an observation back.

        Parameters
        ----------
        X : list of dict-like
            Places where the objective function has already been evaluated.
            Each suggestion is a dictionary where each key corresponds to a
            parameter being optimized.
        y : array-like, shape (n,)
            Corresponding values where objective has been evaluated
        """
        # Update the model with new objective function observations
        # ...
        # No return statement needed
        self._curr_opt.tell(X, y)
        self.switch()
        self.xopt = self._curr_opt.xopt
        self.fopt = self._curr_opt.fopt

    def switch(self):
        """To check if the current optimizer is stopped
        and switch to the next optimizer
        """
        if self._curr_opt.check_stop():
            _max_FEs = self.max_FEs - self._curr_opt.eval_count
            _counter = self._counter + 1

            if _counter < self.N and _max_FEs > 0:
                _curr_opt, _transfer = self.queue[_counter]
                _curr_opt.max_FEs = _max_FEs
                _curr_opt.xopt = np.array(self._curr_opt.xopt)
                _curr_opt.fopt = self._curr_opt.fopt

                if self._transfer:
                    kwargs = self._transfer(self._curr_opt)
                    for k, v in kwargs.items():
                        setattr(_curr_opt, k, v)

                self._logger.name = _curr_opt.__class__.__name__
                self._counter = _counter
                self._curr_opt = _curr_opt
                self._transfer = _transfer
            else:
                self._stop = True
            
    def evaluate(self, X):
        if not hasattr(X[0], '__iter__'):
            return self.obj_fun(X)
        else:
            return [self.obj_fun(x) for x in X]

    def check_stop(self):
        return self._stop


class MultiAcquisitionBO(BO):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        assert self.n_point > 1

        self._acquisition_fun = 'MGFI'
        self._acquisition_fun_list = ['MGFI', 'UCB']
        self._sampler_list = [
            lambda x: np.exp(np.log(x['t']) + 0.5 * np.random.randn()),
            lambda x: 1 / (1 + np.exp((x['alpha'] * 4 - 2) + 0.6 * np.random.randn())) 
        ]
        self._par_name_list = ['t', 'alpha']
        self._acquisition_par_list = [{'t' : 1}, {'alpha' : 0.1}]
        self._N_acquisition = len(self._acquisition_fun_list)

        for i, _n in enumerate(self._par_name_list):
            _criterion = getattr(AcquisitionFunction, self._acquisition_fun_list[i])()
            if _n not in self._acquisition_par_list[i]:
                self._acquisition_par_list[i][_n] = getattr(_criterion, _n)
        
    def _batch_arg_max_acquisition(self, n_point, return_dx):
        criteria = []
        
        for i in range(n_point):
            k = i % self._N_acquisition
            _acquisition_fun = self._acquisition_fun_list[k]
            _acquisition_par = self._acquisition_par_list[k]
            _par = self._sampler_list[k](_acquisition_par)
            _acquisition_par = copy(_acquisition_par)
            _acquisition_par.update({self._par_name_list[k] : _par})
            criteria.append(
                self._create_acquisition(
                    fun=_acquisition_fun, par=_acquisition_par, return_dx=return_dx
                )
            )
        
        if self.n_job > 1:
            __ = Parallel(n_jobs=self.n_job)(
                delayed(self._argmax_restart)(c, logger=self._logger) for c in criteria
            )
        else:
            __ = [list(self._argmax_restart(_, logger=self._logger)) for _ in criteria]
        
        return tuple(zip(*__))


class ParallelBO2(ParallelBO):
    # TODO: add other Parallelization options: 
    # 1) niching-based approach (my EVOLVE paper),
    # 2) bi-objective Pareto-front (PI vs. EI) (my WCCI '16 paper), and
    # 3) maybe QEI?
    pass


class RacingBO(ParallelBO):
    pass


"""
@author: Hao Wang
@email: wangronin@gmail.com
"""
import warnings
import numpy as np

from numpy import sqrt, exp, pi
from scipy.stats import norm
from abc import ABC, abstractmethod

# TODO: implement noisy handling infill criteria, e.g., EQI (expected quantile improvement)
# TODO: perphaps also enable acquisition function engineering here?
# TODO: multi-objective extension

class AcquisitionFunction(ABC):
    def __init__(self, model=None, minimize=True):
        self.model = model
        self.minimize = minimize
    
    @property
    def model(self):
        return self._model
    
    @model.setter
    def model(self, model):
        if model is not None:
            self._model = model
            assert hasattr(self._model, 'predict')
        else: 
            self._model = None

    @abstractmethod
    def __call__(self, X):
        raise NotImplementedError

    def _predict(self, X):
        y_hat, sd2 = self._model.predict(X, eval_MSE=True)
        sd = sqrt(sd2)
        if not self.minimize:
            y_hat = -y_hat
        return y_hat, sd

    def _gradient(self, X):
        y_dx, sd2_dx = self._model.gradient(X)
        if not self.minimize:
            y_dx = -y_dx
        return y_dx, sd2_dx

    def check_X(self, X):
        """Keep input as '2D' object 
        """
        return np.atleast_2d(X)

class ImprovementBased(AcquisitionFunction):
    def  __init__(self, plugin=None, **kwargs):
        super().__init__(**kwargs)
        self.plugin = plugin

    @property
    def plugin(self):
        return self._plugin
    
    @plugin.setter
    def plugin(self, plugin):
        if plugin is None:
            if self._model is not None:
                self._plugin = np.min(self._model.y) if self.minimize \
                    else -1.0 * np.max(self._model.y)
            else:
                self._plugin = None
        else:
            self._plugin = plugin if self.minimize else -1.0 * plugin

class UCB(AcquisitionFunction):
    def __init__(self, alpha=0.5, **kwargs):
        """Upper Confidence Bound 
        """
        super().__init__(**kwargs)
        self.alpha = alpha

    @property
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self, alpha):
        assert alpha >= 0.5 and alpha <= 1
        self._alpha = alpha

    def _predict(self, X):
        y_hat, sd2 = self._model.predict(X, eval_MSE=True)
        sd = sqrt(sd2)

        # NOTE: convert the problem to maximizatio for UCB
        if self.minimize:
            y_hat = -y_hat

        return y_hat, sd

    def _gradient(self, X):
        y_dx, sd2_dx = self._model.gradient(X)
        if self.minimize:
            y_dx = -y_dx
        return y_dx, sd2_dx

    def __call__(self, X, return_dx=False):
        X = self.check_X(X)
        y_, sd_ = self._predict(X)

        try:
            f_value = y_ + sd_ * norm.ppf(self.alpha)
            if len(f_value) == 1:
                f_value = f_value[0]
        except Exception: # in case of numerical errors
            f_value = 0

        if return_dx:
            # TODO: to verify the gradient by rendering it on 2D
            y_dx, sd2_dx = self._gradient(X)
            sd_dx = sd2_dx / (2. * sd_)
            try:
                f_dx = y_dx + sd_dx * norm.ppf(self.alpha)
            except Exception:
                f_dx = np.zeros((len(X[0]), 1))
            return f_value, f_dx

        return f_value

class EI(ImprovementBased):
    def __call__(self, X, return_dx=False):
        X = self.check_X(X)
        n_sample = X.shape[0]
        y_hat, sd = self._predict(X)

        # if the Kriging variance is to small
        # TODO: check the rationale of 1e-6 and why the ratio if intended
        if hasattr(self._model, 'sigma2'):
            if sd / np.sqrt(self._model.sigma2) < 1e-6:
                return (0, np.zeros((len(X[0]), 1))) if return_dx else 0
        else: 
            # TODO: implement a counterpart of 'sigma2' for randomforest
            # or simply put a try...except around the code below
            if sd < 1e-10: 
                return (0, np.zeros((len(X[0]), 1))) if return_dx else 0
        try:
            xcr_ = self._plugin - y_hat
            xcr = xcr_ / sd
            xcr_prob, xcr_dens = norm.cdf(xcr), norm.pdf(xcr)
            f_value = xcr_ * xcr_prob + sd * xcr_dens
            if n_sample == 1:
                f_value = sum(f_value)
        except Exception: # in case of numerical errors
            # IMPORTANT: always keep the output in the same type
            f_value = 0

        if return_dx:
            y_dx, sd2_dx = self._gradient(X)
            sd_dx = sd2_dx / (2. * sd)
            try:
                f_dx = -y_dx * xcr_prob + sd_dx * xcr_dens
            except Exception:
                f_dx = np.zeros((len(X[0]), 1))
            return f_value, f_dx 
        return f_value

class EpsilonPI(ImprovementBased):
    def __init__(self, epsilon=1e-10, **kwargs):
        """epsilon-Probability of Improvement
        # TODO: verify the implementation
        """
        super(EpsilonPI, self).__init__(**kwargs)
        self.epsilon = epsilon
    
    @property
    def epsilon(self):
        return self._epsilon

    @epsilon.setter
    def epsilon(self, eps):
        assert eps > 0
        self._epsilon = eps

    def __call__(self, X, return_dx=False):
        X = self.check_X(X)
        y_hat, sd = self._predict(X)

        coef = 1 - self._epsilon if y_hat > 0 else (1 + self._epsilon)
        try:
            xcr_ = self._plugin - coef * y_hat 
            xcr = xcr_ / sd
            f_value = norm.cdf(xcr)
        except Exception:
            f_value = 0.

        if return_dx:
            y_dx, sd2_dx = self._gradient(X)
            sd_dx = sd2_dx / (2. * sd)
            try:
                f_dx = -(coef * y_dx + xcr * sd_dx) * norm.pdf(xcr) / sd
            except Exception:
                f_dx = np.zeros((len(X[0]), 1))
            return f_value, f_dx 
        return f_value

class PI(EpsilonPI):
    def __init__(self, **kwargs):
        """Probability of Improvement
        """
        kwargs.update({'epsilon' : 0})
        super().__init__(**kwargs)

class MGFI(ImprovementBased):
    def __init__(self, t=1, **kwargs):
        """Moment-Generating Function of Improvement proposed in SMC'17 paper
        """
        super().__init__(**kwargs)
        self.t = t

    @property
    def t(self):
        return self._t

    @t.setter
    def t(self, t):
        assert t > 0
        t = min(t, 22.36)  # bigger `t` value would cause numerical overflow
        self._t = t

    def __call__(self, X, return_dx=False):
        X = self.check_X(X)
        y_hat, sd = self._predict(X)
        n_sample = X.shape[0]

        # if the Kriging variance is to small
        # TODO: check the rationale of 1e-6 and why the ratio if intended
        if np.isclose(sd, 0):
            return (np.array([0.]), np.zeros((len(X[0]), 1))) if return_dx else 0.

        with warnings.catch_warnings():
            warnings.filterwarnings('error')
            try:
                y_hat_p = y_hat - self._t * sd ** 2.
                beta_p = (self._plugin - y_hat_p) / sd
                term = self._t * (self._plugin - y_hat - 1)
                f_ = norm.cdf(beta_p) * exp(term + self._t ** 2. * sd ** 2. / 2.)
                if n_sample == 1:
                    f_ = sum(f_)
            except Exception: # in case of numerical errors
                f_ = 0.

        if np.isinf(f_):
            f_ = 0.
            
        if return_dx:
            y_dx, sd2_dx = self._gradient(X)
            sd_dx = sd2_dx / (2. * sd)

            try:
                term = exp(self._t * (self._plugin + self._t * sd ** 2. / 2 - y_hat - 1))
                m_prime_dx = y_dx - 2. * self._t * sd * sd_dx
                beta_p_dx = -(m_prime_dx + beta_p * sd_dx) / sd
        
                f_dx = term * (norm.pdf(beta_p) * beta_p_dx + \
                    norm.cdf(beta_p) * ((self._t ** 2) * sd * sd_dx - self._t * y_dx))
            except Exception:
                f_dx = np.zeros((len(X[0]), 1))
            return f_, f_dx
        return f_
        
class GEI(ImprovementBased):
    def __init__(self, g=1, **kwargs):
        """Generalized Expected Improvement 
        """
        super().__init__(**kwargs)
        self.g = g

    @property
    def g(self):
        return self._g 
    
    @g.setter
    def g(self, g):
        g = int(g)
        assert g >= 0
        self._g = g

    def __call__(self, X, return_dx=False):
        # TODO: implement this!!!
        raise NotImplementedError


import numpy as np
from typing import Callable, Any, Tuple, List, Union
from scipy.optimize import fmin_l_bfgs_b

from .OnePlusOne_CMA import OnePlusOne_CMA, OnePlusOne_Cholesky_CMA
from .mies import MIES

__all__ = [
    'OnePlusOne_CMA', 'OnePlusOne_Cholesky_CMA', 'MIES', 'argmax_restart'
]

def argmax_restart(
    obj_func: Callable,
    search_space,
    h: Callable = None,
    g: Callable = None,
    eval_budget: int = 100,
    n_restart: int = 10,
    wait_iter: int = 3,
    optimizer: str = 'BFGS',
    logger = None
    ):
    # lists of the best solutions and acquisition values from each restart
    xopt, fopt = [], []  
    best = -np.inf
    wait_count = 0  

    if h is not None or g is not None and optimizer == 'BFGS':
        optimizer = 'OnePlusOne_Cholesky_CMA'

    for iteration in range(n_restart):
        x0 = search_space.sampling(N=1, method='uniform')[0]
        
        # TODO: add constraint handling for BFGS
        if optimizer == 'BFGS':
            mask = np.nonzero(search_space.C_mask | search_space.O_mask)[0]
            bounds = np.array([search_space.bounds[i] for i in mask]) 

            if not all([isinstance(_, float) for _ in x0]):
                raise ValueError('BFGS is not supported with mixed variable types.')

            func = lambda x: tuple(map(lambda x: -1. * x, obj_func(x)))
            xopt_, fopt_, stop_dict = fmin_l_bfgs_b(
                func, x0, pgtol=1e-8, factr=1e6, 
                bounds=bounds, maxfun=eval_budget
            )

            xopt_ = xopt_.flatten().tolist()
            if not isinstance(fopt_, float):
                fopt_ = float(fopt_)
            fopt_ = -fopt_
            
            if logger is not None and stop_dict['warnflag'] != 0:
                logger.debug(
                    'L-BFGS-B terminated abnormally with the state: %s'%stop_dict
                )
        
        elif optimizer == 'OnePlusOne_Cholesky_CMA':
            lb, ub = list(zip(search_space.bounds))

            xopt_, fopt_, stop_dict = OnePlusOne_Cholesky_CMA(
                search_space.dim, obj_func, h=h, g=g,
                lb=lb, ub=ub, max_FEs=eval_budget, 
                minimize=False, verbose=False
            ).run()  
            stop_dict['funcalls'] = stop_dict['FEs']                   

        elif optimizer == 'MIES':
            xopt_, fopt_, stop_dict = MIES(
                search_space, obj_func, eq_func=h, ineq_func=g,
                max_eval=eval_budget, minimize=False, 
                verbose=False, eval_type='list'
            ).optimize()                           

        if fopt_ > best:
            best = fopt_
            wait_count = 0

            if logger is not None:
                logger.debug(
                    'restart : %d - funcalls : %d - Fopt : %f'%(iteration + 1, 
                    stop_dict['funcalls'], fopt_)
                )
        else:
            wait_count += 1

        eval_budget -= stop_dict['funcalls']
        xopt.append(xopt_)
        fopt.append(fopt_)
        
        if eval_budget <= 0 or wait_count >= wait_iter:
            break

    idx = np.argsort(fopt)[::-1]
    return xopt[idx[0]], fopt[idx[0]]


import numpy as np
import atexit, signal, errno, os, sys, time

def arg_to_int(arg):
    if isinstance(arg, str):
        x = int(eval(arg))
    elif isinstance(arg, (int, float)):
        x = int(arg)
    else: 
        raise ValueError
    return x

def dynamic_penalty(X, t, equality=None, inequality=None, C=0.5, alpha=1, beta=2, 
                    epsilon=0.01, minimize=True):
    X = np.atleast_2d(X)
    N = X.shape[0]
    p = np.zeros(N)

    if equality is not None:
        v = np.atleast_2d(list(map(equality, X))).reshape(N, -1)
        v[np.abs(v) <= epsilon] = 0
        p += np.sum(np.abs(v), axis=1)

    if inequality is not None:
        v = np.atleast_2d(list(map(inequality, X))).reshape(N, -1)
        v[v <= 0] = 0
        p += np.sum(np.abs(v) ** beta, axis=1)

    p = (-1) ** (not minimize) * (C * t) ** alpha * p
    return p
    
# TODO: get this done and test it
def stochastic_ranking(X, fitness, equality=None, inquality=None, P=0.4, gamma=1, 
                       beta=1, epsilon=0):
    N = len(X) if isinstance(X, list) else X.shape[0]
    #N = X.shape[0]
    p = np.zeros(N)

    if equality is not None:
        v = np.atleast_2d(list(map(equality, X))).reshape(N, -1)
        v[np.abs(v) <= epsilon] = 0
        p += np.sum(np.abs(v) ** gamma, axis=1)

    if inquality is not None:
        v = np.atleast_2d(list(map(inquality, X))).reshape(N, -1)
        v[v <= 0] = 0
        p += np.sum(np.abs(v) ** beta, axis=1)


# Author: Hao Wang 
# Email: wangronin@gmail.com

from pdb import set_trace
import warnings 

import numpy as np
from numpy.random import uniform
from numpy import log, pi, log10, exp, dot, sqrt, array

from scipy import linalg
from scipy.linalg import cho_solve, solve_triangular, cholesky, LinAlgError
from scipy.optimize import fmin_l_bfgs_b

from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.metrics.pairwise import manhattan_distances
from sklearn.utils import check_random_state, check_array, check_X_y
from sklearn.utils.validation import check_is_fitted

from .cma_es import cma_es
from .trend import constant_trend
from .kernel import absolute_exponential, squared_exponential, generalized_exponential, \
    cubic, pure_nugget, matern
    
from .trend import BasisExpansionTrend, NonparametricTrend

MACHINE_EPSILON = np.finfo(np.double).eps

def l1_cross_distances(X):
    """
    Computes the nonzero componentwise L1 cross-distances between the vectors
    in X.

    Parameters
    ----------

    X: array_like
        An array with shape (n_samples, n_features)

    Returns
    -------

    D: array with shape (n_samples * (n_samples - 1) / 2, n_features)
        The array of componentwise L1 cross-distances.

    ij: arrays with shape (n_samples * (n_samples - 1) / 2, 2)
        The indices i and j of the vectors in X associated to the cross-
        distances in D: D[k] = np.abs(X[ij[k, 0]] - Y[ij[k, 1]]).
    """
    X = check_array(X)
    n_samples, n_features = X.shape
    n_nonzero_cross_dist = n_samples * (n_samples - 1) // 2
    ij = np.zeros((n_nonzero_cross_dist, 2), dtype=np.int)
    D = np.zeros((n_nonzero_cross_dist, n_features))
    ll_1 = 0
    for k in range(n_samples - 1):
        ll_0 = ll_1
        ll_1 = ll_0 + n_samples - k - 1
        ij[ll_0:ll_1, 0] = k
        ij[ll_0:ll_1, 1] = np.arange(k + 1, n_samples)
        D[ll_0:ll_1] = np.abs(X[k] - X[(k + 1):n_samples])

    return D, ij

def my_dot(x, y):
    n_row = x.shape[0]
    n_col = y.shape[1]
    res = np.zeros((n_row, n_col))
    for i in range(n_row):
        for j in range(n_col):
            res[i, j] = np.sum(x[i, :] * y[:, j])
    return res

# TODO: remove the dependences from sklearn
# TODO: simplify this code, which is way too lengthy

class GaussianProcess(BaseEstimator, RegressorMixin):
    """The Gaussian Process model class.

    Read more in the :ref:`User Guide <gaussian_process>`.

    Parameters
    ----------
    mean : string or callable, optional
        A meanession function returning an array of outputs of the linear
        meanession functional basis. The number of observations n_samples
        should be greater than the size p of this basis.
        Default assumes a simple constant meanession trend.
        Available built-in meanession models are::

            'constant', 'linear', 'quadratic'

    corr : string or callable, optional
        A stationary autocorrelation function returning the autocorrelation
        between two points x and x'.
        Default assumes a squared-exponential autocorrelation model.
        Built-in correlation models are::

            'absolute_exponential', 'squared_exponential',
            'generalized_exponential', 'cubic', 'linear', 'matern'

    beta0 : double array_like, optional
        The meanession weight vector to perform Ordinary Kriging (OK).
        Default assumes Universal Kriging (UK) so that the vector beta of
        meanession weights is estimated using the maximum likelihood
        principle.

    storage_mode : string, optional
        A string specifying whether the Cholesky decomposition of the
        correlation matrix should be stored in the class (storage_mode =
        'full') or not (storage_mode = 'light').
        Default assumes storage_mode = 'full', so that the
        Cholesky decomposition of the correlation matrix is stored.
        This might be a useful parameter when one is not interested in the
        MSE and only plan to estimate the BLUP, for which the correlation
        matrix is not required.

    verbose : boolean, optional
        A boolean specifying the verbose level.
        Default is verbose = False.

    theta0 : double array_like, optional
        An array with shape (n_features, ) or (1, ).
        The parameters in the autocorrelation model.
        If thetaL and thetaU are also specified, theta0 is considered as
        the starting point for the maximum likelihood estimation of the
        best set of parameters.
        Default assumes isotropic autocorrelation model with theta0 = 1e-1.

    thetaL : double array_like, optional
        An array with shape matching theta0's.
        Lower bound on the autocorrelation parameters for maximum
        likelihood estimation.
        Default is None, so that it skips maximum likelihood estimation and
        it uses theta0.

    thetaU : double array_like, optional
        An array with shape matching theta0's.
        Upper bound on the autocorrelation parameters for maximum
        likelihood estimation.
        Default is None, so that it skips maximum likelihood estimation and
        it uses theta0.

    normalize : boolean, optional
        Input X and observations y are centered and reduced wrt
        means and standard deviations estimated from the n_samples
        observations provided.
        Default is normalize = True so that data is normalized to ease
        maximum likelihood estimation.

    TODO: the nugget behaves differently than what is described here...
    nugget : double or ndarray, optional
        Introduce a nugget effect to allow smooth predictions from noisy
        data.  If nugget is an ndarray, it must be the same length as the
        number of data points used for the fit.
        The nugget is added to the diagonal of the assumed training covariance;
        in this way it acts as a Tikhonov regularization in the problem.  In
        the special case of the squared exponential correlation function, the
        nugget mathematically represents the variance of the input values.
        Default assumes a nugget close to machine precision for the sake of
        robustness (nugget = 10. * MACHINE_EPSILON).

    optimizer : string, optional
        A string specifying the optimization algorithm to be used.
        Default uses 'fmin_cobyla' algorithm from scipy.optimize.
        Available optimizers are::

            'fmin_cobyla', 'Welch'

        'Welch' optimizer is dued to Welch et al., see reference [WBSWM1992]_.
        It consists in iterating over several one-dimensional optimizations
        instead of running one single multi-dimensional optimization.

    random_start : int, optional
        The number of times the Maximum Likelihood Estimation should be
        performed from a random starting point.
        The first MLE always uses the specified starting point (theta0),
        the next starting points are picked at random according to an
        exponential distribution (log-uniform on [thetaL, thetaU]).
        Default does not use random starting point (random_start = 1).

    random_state: integer or numpy.RandomState, optional
        The generator used to shuffle the sequence of coordinates of theta in
        the Welch optimizer. If an integer is given, it fixes the seed.
        Defaults to the global numpy random number generator.


    Attributes
    ----------
    theta_ : array
        Specified theta OR the best set of autocorrelation parameters (the \
        sought maximizer of the reduced likelihood function).

    log_likelihood_ : array
        The optimal reduced likelihood function value.

    """
    _optimizer_types = ['BFGS', 'CMA']
    _correlation_types = {
        'absolute_exponential': absolute_exponential,
        'squared_exponential': squared_exponential,
        'generalized_exponential': generalized_exponential,
        'cubic': cubic,
        'matern': matern
    }
    _likelihood_functions = ['concentrated', 'restricted']

    # TODO: separater the kernel function from here
    def __init__(
        self, 
        mean=None, 
        corr='squared_exponential',
        theta0=None, thetaL=None, thetaU=None, sigma2=None, 
        nugget=1e-6, noise_estim=False, optimizer='BFGS', 
        likelihood='concentrated', random_start=1, wait_iter=5,
        eval_budget=None, random_state=None, verbose=False
        ):
        self.mean = mean      # Prior mean function 
        self.corr = corr      # Prior correlation function 
        self.sigma2 = sigma2  # variance of the stationary process
        self.verbose = verbose
        self.corr_type = corr
        self.is_fitted = False

        # hyperparameters: kernel function
        self.theta0 = np.array(theta0).flatten() if theta0 is not None else None
        self.thetaL = np.array(thetaL).flatten()
        self.thetaU = np.array(thetaU).flatten()
        
        if not (np.isfinite(self.thetaL).all() and np.isfinite(self.thetaU).all()):
            raise ValueError("all bounds are required finite.")
        
        # hyperparameter optimization parameters
        self.optimizer = optimizer
        self.random_start = random_start
        self.random_state = random_state
        self.wait_iter = wait_iter
        self.eval_budget = eval_budget

        self.noise_var = np.atleast_1d(nugget) if nugget else 0
        self.noise_estim = True if noise_estim else False
        self.noisy = True if self.noise_var or self.noise_estim else False

        # three cases to compute the log-likelihood function
        # TODO: verify: it seems the noisy case is the most useful one
        if not self.noisy:
            self.estimation_mode = 'noiseless'
        elif self.noise_estim:
            self.estimation_mode = 'noise_estim'
        else:
            self.estimation_mode = 'noisy'
        
        assert likelihood in self._likelihood_functions
        self.likelihood = likelihood  # or restricted
        self.is_fitted = False

        if self.mean is None:
            self.mean = constant_trend(len(self.thetaU), beta=0)
        
        # estimation mode for the trend
        if isinstance(self.mean, BasisExpansionTrend):
            self.mean_type = 'basis_expansion'
            self.estimate_trend = True if self.mean.beta is None else False
        elif isinstance(self.mean, NonparametricTrend):
            self.mean_type = 'nonparametric'
        
    def _check_data(self, X, y):
        # Force data to 2D numpy.array
        X, y = check_X_y(X, y, multi_output=False, y_numeric=True)
        y = y.reshape(-1, 1)
        n_samples, _ = X.shape
        self.X, self.y = X, y

        # Run input checks
        self._check_params()

        # Calculate matrix of distances D between samples
        D, ij = l1_cross_distances(X)
        # if (np.min(np.sum(D, axis=1)) == 0. and self.corr != pure_nugget):
        #     raise Exception("Multiple input features cannot have the same"
        #                     " target value.")
        self.D = D
        self.ij = ij

        if self.mean_type == 'basis_expansion' and self.estimate_trend:
            F = self.mean.F(X)
            p = F.shape[1]
            if p > n_samples:
                raise Exception(("Ordinary least squares problem is undetermined "
                                 "n_samples=%d must be greater than the "
                                 "meanession model size p=%d.") % (n_samples, p))
            self.F = F
    
    def sampling_prior(self, X):
        pass
    
    def sampling_posterior(self, X):
        pass
    
    def prior_cov(self, X1, X2=None, corr=False):
        # Check input shapes
        X1 = np.atleast_2d(X1)
        X2 = np.atleast_2d(X2) if X2 else X1

        n_eval_X1, _ = X1.shape
        n_eval_X2, _ = X2.shape
        n_features = self.X.shape[1]
        n_targets = self.y.shape[1]

        if X1.shape[1] != n_features or X2.shape[1] != n_features:
            raise ValueError(("The number of features in X (X.shape[1] = %d) "
                              "should match the number of features used "
                              "for fit() "
                              "which is %d.") % (X1.shape[1], n_features))

        # remember to normalize the inputs
        # if normalize:
            # X1 = (X1 - self.X_mean) / self.X_std
            # X2 = (X2 - self.X_mean) / self.X_std

        dx_new = manhattan_distances(X1, Y=X2, sum_over_features=False)
        R_prior = self.corr(self.theta_, dx_new).reshape(n_eval_X1, n_eval_X2)

        if corr:
            return R_prior
        else:
            C_prior = array([self.sigma2[i] * R_prior for i in range(n_targets)])
            C_prior = sqrt((C_prior ** 2.).sum(axis=0) / n_targets)

            return C_prior

    def fit(self, X, y):
        """The Gaussian Process model fitting method.

        Parameters
        ----------
        X : double array_like
            An array with shape (n_samples, n_features) with the input at which
            observations were made.

        y : double array_like
            An array with shape (n_samples, ) or shape (n_samples, n_targets)
            with the observations of the output to be predicted.

        Returns
        -------
        gp : self
            A fitted Gaussian Process model object awaiting data to perform
            predictions.
        """
        # Run input checks
        self.random_state = check_random_state(self.random_state)
        self._check_data(X, y)

        # Determine Gaussian Process model parameters
        if self.thetaL is not None and self.thetaU is not None:
            # Maximum Likelihood Estimation of the parameters
            if self.verbose:
                print("Maximum Likelihood Estimation of the hyperparameters...")

            while True:
                self.par, self.log_likelihood_, env = self._optimize_hyperparameter()
                if np.isinf(self.log_likelihood_):
                    # TODO: open design choice here:
                    #   1) leave this exception handeling part to the program calls gpr
                    #   2) handel it here and save the incident to a log
                    print('Invalid likelihood value. Increasing nugget...')

                    # TODO: maybe turn the working mode to 'noise_estim' directly
                    if self.estimation_mode == 'noiseless':
                        self.estimation_mode = 'noisy'
                        self.noise_var = 1e-5
                    else:
                        self.noise_var *= 10
                else:
                    break

        # find a better name for noise_var
        self.theta_ = self.par['theta']
        self.noise_var = env['noise_var']
        self.sigma2 = env['sigma2']
        self.rho = env['rho'] 
        self.Yt = env['Yt']
        self.C = env['C']
        if self.mean_type == 'basis_expansion' and self.estimate_trend:
            self.Ft = env['Ft']
            self.G = env['G']
            self.Q = env['Q']

        # compute for beta and gamma
        self.compute_beta_gamma()
        self.is_fitted = True
        return self

    def update(self, X, y):
        # TODO: implement incremental training 
        self.fit(X, y)
        return self

    def predict(self, X, eval_MSE=False, batch_size=None):
        """
        This function evaluates the Gaussian Process model at x.

        Parameters
        ----------
        X : array_like
            An array with shape (n_eval, n_features) giving the point(s) at
            which the prediction(s) should be made.

        eval_MSE : boolean, optional
            A boolean specifying whether the Mean Squared Error should be
            evaluated or not.
            Default assumes evalMSE = False and evaluates only the BLUP (mean
            prediction).

        batch_size : integer, optional
            An integer giving the maximum number of points that can be
            evaluated simultaneously (depending on the available memory).
            Default is None so that all given points are evaluated at the same
            time.

        Returns
        -------
        y : array_like, shape (n_samples, ) or (n_samples, n_targets)
            An array with shape (n_eval, ) if the Gaussian Process was trained
            on an array of shape (n_samples, ) or an array with shape
            (n_eval, n_targets) if the Gaussian Process was trained on an array
            of shape (n_samples, n_targets) with the Best Linear Unbiased
            Prediction at x.

        MSE : array_like, optional (if eval_MSE == True)
            An array with shape (n_eval, ) or (n_eval, n_targets) as with y,
            with the Mean Squared Error at x.
        """
        assert hasattr(self, 'X')

        # Check input shapes
        # TODO: remove the support for multiple independent outputs
        X = check_array(X)
        n_eval, _ = X.shape
        n_samples, n_features = self.X.shape
        n_targets = self.y.shape[1]

        # Run input checks
        self._check_params()

        if X.shape[1] != n_features:
            raise ValueError(("The number of features in X (X.shape[1] = %d) "
                              "should match the number of features used "
                              "for fit() which is %d.") % (X.shape[1], n_features))

        if batch_size is None:
            # (evaluates all given points in a single batch run)
            # Initialize output
            y = np.zeros(n_eval)
            if eval_MSE:
                MSE = np.zeros(n_eval)

            # Get pairwise componentwise L1-distances to the input training set
            # TODO: remove calculations of distances from here
            dx = manhattan_distances(X, Y=self.X, sum_over_features=False)
            # Get meanession function and correlation
            r = self.corr(self.theta_, dx).reshape(n_eval, n_samples)

            # Predictor
            # TODO: get rid of n_targets in this class
            y = (self.mean(X) + my_dot(r, self.gamma)).reshape(n_eval, n_targets).ravel()
            
            # Mean Squared Error
            if eval_MSE:
                rt = solve_triangular(self.C, r.T, lower=True)

                if self.estimate_trend:  # Universal / Ordinary Kriging
                    f = self.mean.F(X)
                    u = solve_triangular(self.G.T, np.dot(self.Ft.T, rt) - f.T, lower=True)
                else:                  # simple Kriging
                    u = np.zeros((n_targets, n_eval))

                MSE = np.dot(self.sigma2.reshape(n_targets, 1),
                             (1. - (rt ** 2.).sum(axis=0) + 
                             (u ** 2.).sum(axis=0))[np.newaxis, :])
                MSE = np.sqrt((MSE ** 2.).sum(axis=0) / n_targets)

                # Mean Squared Error might be slightly negative depending on
                # machine precision: force to zero!
                MSE[MSE < 0.] = 0.
                MSE = MSE.ravel()

                return y, MSE
            else:
                return y
        else:
            # Memory management
            if type(batch_size) is not int or batch_size <= 0:
                raise Exception("batch_size must be a positive integer")

            if eval_MSE:
                y, MSE = np.zeros(n_eval), np.zeros(n_eval)
                for k in range(max(1, n_eval / batch_size)):
                    batch_from = k * batch_size
                    batch_to = min([(k + 1) * batch_size + 1, n_eval + 1])
                    y[batch_from:batch_to], MSE[batch_from:batch_to] = \
                        self.predict(X[batch_from:batch_to],
                                     eval_MSE=eval_MSE, batch_size=None)

                return y, MSE

            else:

                y = np.zeros(n_eval)
                for k in range(max(1, n_eval / batch_size)):
                    batch_from = k * batch_size
                    batch_to = min([(k + 1) * batch_size + 1, n_eval + 1])
                    y[batch_from:batch_to] = \
                        self.predict(X[batch_from:batch_to],
                                     eval_MSE=eval_MSE, batch_size=None)
                return y

    def gradient(self, x):
        """Calculate the gradient of the posterior mean and variance
        Note that the nugget effect will not the change the computation below
        """
        x = np.atleast_2d(x)
        n_eval, _ = x.shape
        n_samples, n_features = self.X.shape

        if _ != n_features:
            raise Exception('x does not have the right size!')

        if n_eval != 1:
            raise Exception('x must be a vector!')

        # trend and its Jacobian
        f = self.mean.F(x).reshape(-1, 1)
        f_dx = self.mean.Jacobian(x)
        
        # correlation and its Jacobian
        d = manhattan_distances(x, Y=self.X, sum_over_features=False)
        r = self.corr(self.theta_, d).reshape(n_eval, n_samples)
        r_dx = self.corr_dx(x, X=self.X, r=r).T
        
        # gradient of the posterior mean
        y_dx = dot(self.mean.beta.T, f_dx) + my_dot(self.gamma.T, r_dx)

        # auxiliary variable: rt = C^-1 * r
        rt = solve_triangular(self.C, r.T, lower=True)
        rt_dx = solve_triangular(self.C, r_dx, lower=True)

        mse_dx = -1.0 * dot(rt.T, rt_dx)      # for Simple Kriging
        if self.estimate_trend:       # Universal / Ordinary Kriging
            # auxiliary variable: u = Ft^T * rt - f
            u = dot(self.Ft.T, rt) - f
            u_dx = dot(self.Ft.T, rt_dx) - f_dx
            Ft2inv = linalg.inv(dot(self.Ft.T, self.Ft))
            mse_dx += u.T.dot(Ft2inv).dot(u_dx)

        mse_dx = 2.0 * self.sigma2 * mse_dx
        return y_dx.T, mse_dx.T

    def Hessian(self, x):
        """Calculate the Hessian matrix of the posterior mean at the input point `x`
        """
        x = np.atleast_2d(x)
        n_eval, _ = x.shape
        n_samples, n_features = self.X.shape

        if _ != n_features:
            raise Exception('x does not have the right size!')

        if n_eval != 1:
            raise Exception('x must be a vector!')

        # The Hessian tensor of the trend
        f_dx2 = self.mean.Hessian(x)
        
        # The Hessian tensor of the correlation
        d = manhattan_distances(x, Y=self.X, sum_over_features=False)
        r = self.corr(self.theta_, d).reshape(n_eval, n_samples)
        r_dx2 = self.corr_Hessian(x, X=self.X, r=r)
        
        return (f_dx2.dot(self.mean.beta) + r_dx2.dot(self.gamma))[..., 0]

    def corr_dx(self, x, X=None, theta=None, r=None, nu=1.5):
        # TODO: move corr_dx, corr_grad_theta the kernel module
        # TODO: rename corr_grad_theta to something else
        # Check input shapes
        x = np.atleast_2d(x)
        n_eval, _ = x.shape
        n_samples, n_features = self.X.shape

        if _ != n_features:
            raise Exception('x does not have the right size!')

        if n_eval != 1:
            raise Exception('x must be a vector!')

        if self.theta_ is None:
            raise Exception('The model is not fitted yet!')

        X = np.atleast_2d(X)
        if X is None:
            X = self.X

        diff = (x - X).T

        if theta is None:
            theta = self.theta_

        # calculate the required variables if not given
        if r is None:
            r = self.corr(self.theta_, np.abs(diff).T)

        theta = theta.reshape(-1, 1)

        with warnings.catch_warnings():
            warnings.filterwarnings('error')
            try:
                if self.corr_type == 'squared_exponential':
                    grad = -2 * r * (theta * diff)

                elif self.corr_type == 'matern':
                    c = np.sqrt(3)
                    D = np.sqrt(np.sum(theta * diff ** 2., axis=0))

                    if nu == 0.5:
                        grad = - diff * theta / D * r
                    elif nu == 1.5:
                        grad = diff * theta / D
                        grad *= -3. * D * exp(-c * D)
                    elif nu == 2.5:
                        pass

                elif self.corr_type == 'absolute_exponential':
                    grad = -1.0 * r * theta * np.sign(diff)
                elif self.corr_type == 'generalized_exponential':
                    pass
                elif self.corr_type == 'cubic':
                    pass
                elif self.corr_type == 'linear':
                    pass
            except Warning:
                grad = np.zeros((n_features, n_samples))

        return grad
    
    def corr_Hessian(self, x, X=None, theta=None, r=None, nu=1.5):
        # Check input shapes
        x = np.atleast_2d(x)
        n_eval, _ = x.shape
        n_samples, n_features = self.X.shape
        assert n_eval == 1

        if _ != n_features:
            raise Exception('x does not have the right size!')

        if n_eval != 1:
            raise Exception('x must be a vector!')

        if self.theta_ is None:
            raise Exception('The model is not fitted yet!')

        X = np.atleast_2d(X)
        if X is None:
            X = self.X

        diff = (x - X).T

        if theta is None:
            theta = self.theta_
        theta = theta.reshape(-1, 1)

        # calculate the required variables if not given
        if r is None:
            r = self.corr(self.theta_, np.abs(diff).T)

        with warnings.catch_warnings():
            warnings.filterwarnings('error')
            try:
                if self.corr_type == 'squared_exponential':
                    diff_ = theta * diff
                    g = -2 * r * diff_

                    H = []
                    for k in range(n_features):
                        _ = np.zeros((n_features, n_samples))
                        _[k, :] = theta[k]
                        H.append(
                            -2 * (g[k, :] * (theta * diff) + r * _)
                        )
                    H = np.atleast_3d(H)
                elif self.corr_type == 'matern':
                    c = np.sqrt(3)
                    D = np.sqrt(np.sum(theta * diff ** 2., axis=0))

                    if nu == 0.5:
                        grad = - diff * theta / D * r
                    elif nu == 1.5:
                        grad = diff * theta / D
                        grad *= -3. * D * exp(-c * D)
                    elif nu == 2.5:
                        pass

                elif self.corr_type == 'absolute_exponential':
                    diff_ = theta * np.sign(diff)
                    g = -1.0 * r * diff_
                    H = np.atleast_3d([
                        r[i] * np.diag(theta * diff[:, i] / np.sign(diff[:, i])) + \
                            np.tile(g, (1, n_features)) * diff_[:, i] for i in range(n_samples)
                        ]
                    )
                    H *= -1
                elif self.corr_type == 'generalized_exponential':
                    pass
            except Warning:
                H = np.zeros((n_features, n_samples))

        return H

    def corr_grad_theta(self, theta, X, R, nu=1.5):
        # Check input shapes
        X = np.atleast_2d(X)
        _ = X.shape[1]
        n_features = self.X.shape[1]

        if _ != n_features:
            raise Exception('x does not have the right size!')

        diff = (X[:, np.newaxis, :] - X[np.newaxis, :, :]) ** 2.

        if self.corr_type == 'squared_exponential':
            grad = -diff * R[..., np.newaxis]

        elif self.corr_type == 'matern':
            c = np.sqrt(3)
            D = np.sqrt(np.sum(theta * diff, axis=-1))

            if nu == 0.5:
                grad = - diff * theta / D * R
            elif nu == 1.5:
                grad = -3 * np.exp(-c * D)[..., np.newaxis] * diff / 2.
            elif nu == 2.5:
                pass

        elif self.corr_type == 'absolute_exponential':
            grad = -np.sqrt(diff) * R[..., np.newaxis]
        elif self.corr_type == 'generalized_exponential':
            pass
        elif self.corr_type == 'cubic':
            pass
        elif self.corr_type == 'linear':
            pass

        return grad

    def correlation_matrix(self, theta, X=None):
        D = self.D
        ij = self.ij
        n_samples = self.X.shape[0]
        
        # calculate the correlation matrix R
        r = self.corr(theta, D)
        R = np.eye(n_samples)
        R[ij[:, 0], ij[:, 1]] = r
        R[ij[:, 1], ij[:, 0]] = r
        return R

    def compute_beta_gamma(self):
        if self.estimate_trend:
            # estimate the trend coefficients
            self.mean.beta = solve_triangular(self.G, my_dot(self.Q.T, self.Yt))
        self.gamma = solve_triangular(self.C.T, self.rho).reshape(-1, 1)

    def _compute_aux_var(self, R):
        # compute auxiliary variables
        # Cholesky decomposition of R: Note that this matrix R can be singular
        # change notation from 'C' to 'L': 'L'ower triagular component...
        try:
            L = cholesky(R, lower=True)
        except LinAlgError as e:
            raise e

        Yt = solve_triangular(L, self.y, lower=True)
        if self.estimate_trend:
            # compute rho: the residuals in linear model
            # ρ = Y - Fβ 
            # both Y and F are transformed by R^-1/2 such that the noise covariance is identity
            Ft = solve_triangular(L, self.F, lower=True)
            Q, G = linalg.qr(Ft, mode='economic')
            rho = Yt - Q.dot(Q.T).dot(Yt)
        else:
            rho = Yt - solve_triangular(L, self.mean(self.X), lower=True)
            Ft, Q, G = None, None, None

        return L, Ft, Yt, Q, G, rho
    
    def log_likelihood_restricted(self, par, env=None, eval_grad=False):
        """
        TODO: write this function in C or Cython
        The restricted log likelihood function
        It yields the unbiased estimation for the hyperparameters 
        """
        # TODO: find a way to replace this function
        # check_is_fitted(self, "X")
        n_samples = self.X.shape[0]
        n_par = len(par)
        log_likelihood = -np.inf
        
        # the MLE estimation of sigma2 is not used here as it will make estimation of theta biased
        if self.estimation_mode == 'noiseless':
            theta, sigma2, noise_var = par[:-1], par[-1], 0
            
        elif self.estimation_mode == 'noisy':
            theta, sigma2, noise_var = par[:-1], par[-1], self.noise_var
            
        elif self.estimation_mode == 'noise_estim':
            theta, sigma2, noise_var = par[:-2], par[-2], par[-1]
            
        R0 = self.correlation_matrix(theta)
        total_var = sigma2 + noise_var
        C = sigma2 * R0 + noise_var * np.eye(n_samples)  # homoscedastic noises are assumed
        R = C / total_var
            
        try:
            L, Ft, Yt, Q, G, rho = self._compute_aux_var(R)
        except LinAlgError as e:
            warnings.warn("linear operation failed on {}".format(e))
            if eval_grad:
                return log_likelihood, np.zeros((n_par, 1))
            else:
                return log_likelihood
            
        if self.estimate_trend: 
            p = Ft.shape[1]
            log_likelihood = -0.5 * ((n_samples - p) * log(2 * pi * total_var) -
                                     np.log(np.linalg.det(self.F.T.dot(self.F))) + 
                                     2 * np.log(np.diag(L)).sum() + 
                                     np.log(np.diag(G).prod() ** 2) +
                                     rho.T.dot(rho) / total_var).sum()
        else:  # simple kriging
            log_likelihood = -0.5 * (n_samples * log(2 * pi * total_var) - 
                                     2 * np.log(np.diag(L)).sum() + 
                                     rho.T.dot(rho) / total_var).sum()
            
        if np.exp(log_likelihood) > 1:
            if self.verbose:
                warnings.warn("invalid log likelihood value: {}".format(log_likelihood))
            log_likelihood = -np.inf

        if eval_grad:
            # gradient calculation of the log-likelihood
            gamma_ = solve_triangular(L.T, rho).reshape(-1, 1) / total_var
            Cinv = cho_solve((L, True), np.eye(n_samples)) / total_var
            if self.estimate_trend:
                __ = solve_triangular(L.T, Q)
                term = __.dot(__.T)
    
            # Covariance: partial derivatives w.r.t. theta
            C_grad_tensor = total_var * self.corr_grad_theta(theta, self.X, R0)
            # Covariance: partial derivatives w.r.t. sigma2
            C_grad_tensor = np.concatenate([C_grad_tensor, R0[..., np.newaxis]], axis=2)
            
            if self.estimation_mode == 'noise_estim':
                # Covariance: partial derivatives w.r.t. noise_var
                C_grad_tensor = np.concatenate([C_grad_tensor, np.eye(n_samples)[..., np.newaxis]], 
                                                axis=2)
            # partial dderivatives of llf
            gradient = np.zeros((n_par, 1))
            for i in range(n_par):
                C_grad = C_grad_tensor[:, :, i]
                # TODO: beta hat is not considered in the partial derivatives
                if self.estimate_trend: 
                    gradient[i] = -0.5 * (np.sum(Cinv * C_grad) - 
                                          gamma_.T.dot(C_grad).dot(gamma_) - 
                                          np.sum(term * C_grad))
                else:  # simple kriging
                    gradient[i] = -0.5 * (np.sum(Cinv * C_grad) - 
                                          gamma_.T.dot(C_grad).dot(gamma_))
                    
        if env is not None:
            env['sigma2'] = sigma2
            env['noise_var'] = noise_var
            env['rho'] = rho
            env['Yt'] = Yt
            env['C'] = L
            if self.estimate_trend:
                env['Ft'] = Ft 
                env['G'] = G
                env['Q'] = Q
                
        if eval_grad:
            return log_likelihood, gradient
        else:
            return log_likelihood

    def log_likelihood_concentrated(self, par, env=None, eval_grad=False):
        """The concentrated log likelihood function, which underestimates `sigma2`
        
        Parameters that are concetrated out:
            beta : coeffiencts in the basis expansion trend function
                replaced by the MLE estimate (GLS formula) in the likelihood 
            sigma2 : stationary variance of the process
                replace by the MLE estimate
        """
        n_samples = self.X.shape[0]
        n_par = len(par)
        log_likelihood = -np.inf
        if self.estimation_mode == 'noiseless':
            theta = par
            noise_var = 0

            R0 = self.correlation_matrix(theta)
            
            try:
                L, Ft, Yt, Q, G, rho = self._compute_aux_var(R0)
            except LinAlgError:
                if eval_grad:
                    return (log_likelihood, np.zeros((n_par, 1)))
                else:
                    return log_likelihood
            
            # TODO: experimental correction of the sigma2 estimation
            if Q is not None:
                k = np.linalg.matrix_rank(Q.dot(Q.T))
            else:
                k = 0
            sigma2 = (rho ** 2.).sum(axis=0) / (n_samples - k)
            log_likelihood = -0.5 * (n_samples * log(2. * pi * sigma2) + 
                                     2. * np.log(np.diag(L)).sum() + n_samples)

        elif self.estimation_mode == 'noise_estim':
            theta, alpha = par[:-1], par[-1]
            R0 = self.correlation_matrix(theta)
            R = alpha * R0 + (1 - alpha) * np.eye(n_samples)

            try:
                L, Ft, Yt, Q, G, rho = self._compute_aux_var(R)
            except LinAlgError:
                if eval_grad:
                    return (log_likelihood, np.zeros(n_par, 1))
                else:
                    return log_likelihood
                
            sigma2_total = (rho ** 2.).sum(axis=0) / n_samples
            sigma2, noise_var = alpha * sigma2_total, (1 - alpha) * sigma2_total
            log_likelihood = -0.5 * (n_samples * log(2. * pi * sigma2_total) + 
                                     2. * np.log(np.diag(L)).sum() + n_samples)

        elif self.estimation_mode == 'noisy':
            theta, sigma2 = par[:-1], par[-1]
            noise_var = self.noise_var
            sigma2_total = sigma2 + noise_var

            R0 = self.correlation_matrix(theta)
            C = sigma2 * R0 + noise_var * np.eye(n_samples)
            R = C / sigma2_total

            try:
                L, Ft, Yt, Q, G, rho = self._compute_aux_var(R)
            except LinAlgError:
                if eval_grad:
                    return (log_likelihood, np.zeros(n_par, 1)) 
                else: 
                    return log_likelihood

            log_likelihood = -0.5 * (n_samples * log(2. * pi * sigma2_total) + 
                                     2. * np.log(np.diag(L)).sum() + 
                                     np.dot(rho.T, rho) / sigma2_total)

        if env is not None:
            env['sigma2'] = sigma2
            env['noise_var'] = noise_var
            env['rho'] = rho
            env['Yt'] = Yt
            # TODO: change variable 'C' --> 'L'
            env['C'] = L
            env['Ft'] = Ft
            env['G'] = G
            env['Q'] = Q

        # for verificationn
        # TODO: remove this in the future
        if log_likelihood > 0:
            # warnings.warn("invalid log likelihood value: {}".format(log_likelihood))
            return (-np.inf, np.zeros((n_par, 1))) if eval_grad else -np.inf
            
        if not eval_grad:
            return log_likelihood[0]

        # gradient calculation of the log-likelihood
        gamma = solve_triangular(L.T, rho).reshape(-1, 1)
        
        Rinv = cho_solve((L, True), np.eye(n_samples))
        Rinv_upper = Rinv[np.triu_indices(n_samples, 1)]
        _upper = gamma.dot(gamma.T)[np.triu_indices(n_samples, 1)]

        llf_grad = np.zeros((n_par, 1))

        if self.estimation_mode == 'noiseless':
            # The grad tensor of R w.r.t. theta
            R_grad_tensor = self.corr_grad_theta(theta, self.X, R0)

            for i in range(n_par):
                R_grad_upper = R_grad_tensor[:, :, i][np.triu_indices(n_samples, 1)]

                llf_grad[i] = np.sum(_upper * R_grad_upper) / sigma2 \
                    - np.sum(Rinv_upper * R_grad_upper)

        elif self.estimation_mode == 'noise_estim':
            # The grad tensor of R w.r.t. theta
            R_grad_tensor = alpha * self.corr_grad_theta(theta, self.X, R0)

            # partial derivatives w.r.t theta's
            for i in range(n_par - 1):
                R_grad_upper = R_grad_tensor[:, :, i][np.triu_indices(n_samples, 1)]

                # Note that sigma2_total is used here
                llf_grad[i] = np.sum(_upper * R_grad_upper) / sigma2_total \
                    - np.sum(Rinv_upper * R_grad_upper)

            # partial derivatives w.r.t 'v'
            R_dv = R0 - np.eye(n_samples)
            llf_grad[n_par - 1] = -0.5 * (np.sum(Rinv * R_dv) - 
                                          np.dot(gamma.T, R_dv.dot(gamma)) / sigma2_total)

        elif self.estimation_mode == 'noisy':
            gamma_ = gamma / sigma2_total
            Cinv = Rinv / sigma2_total
            # Covariance: partial derivatives w.r.t. theta
            C_grad_tensor = sigma2_total * self.corr_grad_theta(theta, self.X, R0)
            
            # Covariance: partial derivatives w.r.t. sigma2
            C_grad_tensor = np.concatenate([C_grad_tensor, R0[..., np.newaxis]], axis=2)

            for i in range(n_par):
                C_grad = C_grad_tensor[:, :, i]
                llf_grad[i] = -0.5 * (np.sum(Cinv * C_grad) - 
                                      np.dot(gamma_.T, C_grad).dot(gamma_))
                
        return log_likelihood[0], llf_grad
    
    def _hyperparameter_bound(self, par_list):
        bounds = []
        # TODO: better estimation the upper and lowe bound of sigma2
        for name in par_list:
            if name == 'theta':
                bounds.append(np.c_[self.thetaL, self.thetaU])
            elif name == 'sigma2':
                bounds.append(np.atleast_2d([1e-5, self.y.std() ** 2]))
            elif name == 'alpha':
                bounds.append(np.atleast_2d([1e-10, 1.0 - 1e-10]))
            elif name == 'noise_var':
                # TODO: implement this
                bounds.append(np.atleast_2d([1e-10, 1.0 - 1e-10]))
        bounds = np.concatenate(bounds, axis=0)
        return bounds
        
    def _optimize_hyperparameter(self):
        """
        optimization procedure to determine the hyperparameter:
            theta: corralation length
            sigma2: process variance for stationary GP
            noise_var: the variance(s) of the noise process
        -------
        """

        if self.verbose:
            print("The chosen optimizer is: " + str(self.optimizer))
            print('Estimation mode: {}'.format(self.estimation_mode))
            print('Likelihood function: {}'.format(self.likelihood))
            print("{} random restarts are specified.".format(self.random_start))
        
        par_list = ['theta']
        par_len = [len(self.thetaL)]
        if self.likelihood == 'restricted' or self.estimation_mode == 'noisy':
            # TODO: implement optimization for the heterogenous case
            par_list += ['sigma2']
            par_len.append(1)
        
        if self.estimation_mode == 'noise_estim':
            if self.likelihood == 'concentrated':
                par_list += ['alpha']
                par_len.append(1)
            elif self.likelihood == 'restricted':
                par_list += ['noise_var']
                par_len.append(1)
        
        bounds = self._hyperparameter_bound(par_list)
        log10bounds = log10(bounds)
        n_theta = len(self.thetaL)
        # if the model has been optimized before as the starting point,
        # then use the last optimized hyperparameters
        # supposed to be good for updating the model incrementally
        # TODO: validate this
        if hasattr(self, 'theta_'):
            log10theta0 = log10(self.theta_)
        else:
            log10theta0 = log10(self.theta0) if self.theta0 is not None else \
                np.random.uniform(log10(bounds)[:, 0], log10(bounds)[:, 1])

        if self.estimation_mode == 'noiseless' and self.likelihood == 'concentrated':
            log10param = log10theta0
        else:
            log10param = np.r_[log10theta0, uniform(log10bounds[n_theta:, 0], 
                                                    log10bounds[n_theta:, 1])]
            
        n_par = len(log10param)
        eval_budget = 200 * n_par if self.eval_budget is None else self.eval_budget
        llf_opt = np.inf
        
        def _obj_func(eval_grad=False):
            def func(log10param):
                self.eval_count += 1
                param = 10. ** np.array(log10param)
                if self.likelihood == 'concentrated':
                    __ = self.log_likelihood_concentrated(param, eval_grad=eval_grad)
                elif self.likelihood == 'restricted':
                    __ = self.log_likelihood_restricted(param, eval_grad=eval_grad)
                return tuple(-_ for _ in tuple(__))
            return func
        
        self.eval_count = 0
        # TODO: maybe adopt an ILS-like restarting heuristic?
        if self.optimizer == 'BFGS':            
            obj_func = _obj_func(eval_grad=True)
            wait_count = 0  # stagnation counter
            
            for iteration in range(self.random_start):
                if iteration != 0:
                    log10param = np.random.uniform(log10bounds[:, 0], log10bounds[:, 1])
                
                # TODO: may be expose the parameter of fmin_l_bfgs_b to the user
                param_opt_, llf_opt_, info = fmin_l_bfgs_b(obj_func, log10param, 
                                                           bounds=log10bounds,
                                                           maxfun=eval_budget)
                
                # TODO: verify this rule to determine the marginal improvement 
                # diff = (llf_opt - llf_opt_) / max(abs(llf_opt_), abs(llf_opt), 1)
                if iteration == 0:
                    param_opt = param_opt_
                    llf_opt = llf_opt_    
                # elif diff >= 1e7 * MACHINE_EPSILON:
                elif llf_opt_ <= llf_opt:
                    param_opt, llf_opt = param_opt_, llf_opt_
                    wait_count = 0
                else:
                    wait_count += 1

                if self.verbose:
                    print('restart {} takes {} evals'.format(iteration + 1, info['funcalls']))
                    print('best log likekihood value: {}'.format(-llf_opt))
                    if info["warnflag"] != 0:
                        warnings.warn("fmin_l_bfgs_b terminated abnormally with "
                                      "the state: {}".format(info))
                                      
                eval_budget -= info['funcalls']
                if eval_budget <= 0 or wait_count >= self.wait_iter:
                    break
        
        elif self.optimizer == 'CMA':   # IPOP-CMA-ES 
            obj_func = _obj_func()
            opt = {'sigma_init': 0.25 * np.max(log10bounds[:, 1] - log10bounds[:, 0]),
                   'eval_budget': eval_budget,
                   'f_target': np.inf,
                   'lb': log10bounds[:, 1],
                   'ub': log10bounds[:, 0],
                   'restart_budget': self.random_start}
            
            # TODO: perphas use the BIPOP-CMA-ES in the future     
            optimizer = cma_es(n_par, log10param, obj_func, opt, is_minimize=False, restart='IPOP')
            param_opt, llf_opt, evalcount, info = optimizer.optimize()
            param_opt = param_opt.flatten()
            
            if self.verbose:
                print('{} evals, best log likekihood value: {}'.format(evalcount, -llf_opt))
        
        optimal_param = 10. ** param_opt
        env = {}
        if self.likelihood == 'concentrated':
            optimal_llf_value = self.log_likelihood_concentrated(optimal_param, env)
        elif self.likelihood == 'restricted':
            optimal_llf_value = self.log_likelihood_restricted(optimal_param, env)
        
        param = {}
        i = 0
        for k, name in enumerate(par_list):
            len_ = par_len[k]
            param[name] = optimal_param[i:i+len_]
            i += len_

        return param, optimal_llf_value, env

    def _check_params(self):
        # Check correlation model
        if not callable(self.corr):
            if self.corr in self._correlation_types:
                self.corr = self._correlation_types[self.corr]
            else:
                raise ValueError("corr should be one of %s or callable, "
                                 "%s was given." % (self._correlation_types.keys(), self.corr))

        # Check correlation parameters
        # self.theta0 = np.atleast_2d(self.theta0)
        # lth = self.theta0.size

        if self.thetaL is not None and self.thetaU is not None:
            if self.thetaL.size != self.thetaU.size:
                raise ValueError("thetaL and thetaU must have the same length.")
            if self.theta0 is not None and self.theta0.size != self.thetaL.size:
                raise ValueError("theta0, thetaL, and thetaU must have the same length.")
            if np.any(self.thetaL <= 0) or np.any(self.thetaU < self.thetaL):
                raise ValueError("The bounds must satisfy O < thetaL <= "
                                 "thetaU.")

        elif self.thetaL is None and self.thetaU is None:
            if self.theta0 is None:
                raise ValueError("theta0 must be provided when both thetaL and thetaU are not set.")
            if np.any(self.theta0 <= 0):
                raise ValueError("theta0 must be strictly positive.")

        elif self.thetaL is None or self.thetaU is None:
            raise ValueError("thetaL and thetaU should either be both or "
                             "neither specified.")

        # Force verbose type to bool
        self.verbose = bool(self.verbose)

        # Check nugget value
        # self.nugget = np.asarray(self.nugget)
        # if np.any(self.nugget) < 0.:
        #     raise ValueError("nugget must be positive or zero.")
        # if (n_samples is not None
        #         and self.nugget.shape not in [(), (n_samples,)]):
        #     raise ValueError("nugget must be either a scalar "
        #                      "or array of length n_samples.")

        # Check optimizer
        if self.optimizer not in self._optimizer_types:
            raise ValueError("optimizer should be one of %s" % self._optimizer_types)

        # Force random_start type to int
        self.random_start = int(self.random_start)
